/***************************************************************************
 *
 * This is a part of AntMon project
 *
 *            ant_act.cpp
 *
 *  Tue Aug 22 14:33:28 2006
 *  Copyright  2006  Sergey Zhumatiy
 *  serg@parallel.ru
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>

#include "ant_act.hpp"
#include "ant_log.hpp"
#include "ant_exe.hpp"
/*#include "ant_hashes.hpp"*/

void gentle_exit(int code);

#define DEBUG

using namespace std;

int Ant_action::send_retr=20; /* 10 seconds */
time_t Ant_action::reinit_interval=600; /* 10 minuites */

struct timeval Ant_action::sleep_time={0,500000};
map<string, int *> Ant_action::global_exe;
map<string, int *> Ant_action::global_err;

typedef map<string, Ant_action *>::iterator hi;

string Ant_action::actions_path=DEF_EXE_PATH;

extern time_t global_time;

//
//  Write message to buffer
//
void Ant_action::sendMessage(const char *m){
  int len;

  if(!active || !fd){
    if(global_time-time_disabled>reinit_interval)
      reinit();

    if(!active || !fd)
      return;
  }

  len=strlen(m);
  while(len+pos>=size){
    // resize buffer!

    size+=ANT_ACT_BUF_CHUNK;
    char *buf2=new char[size];
    if(buf2==NULL){
      LOG(ALOG_WARN,("Cannot allocate %d bytes of memory.",size));
      gentle_exit(10);
    }

    // copy old buffer to new
    memcpy(buf2,buffer,size-ANT_ACT_BUF_CHUNK);
    {/*log("del 01"); */delete[] buffer;}
    buffer=buf2;
  }// resize

  // copy message
  memcpy(buffer+pos,m,len+1);
  pos+=len;
}


int Ant_action::flush(){
  int rest=pos;
  int sent,retries,fail;
  struct timeval sl;

  if(!active || !fd){
    if(global_time-time_disabled>reinit_interval)
      reinit();

    if(!active || !fd)
      return 0;
  }
  if(rest==0)
   return 0;

  retries=send_retr+1;

  if(rest){LOG(ALOG_DEBUG2,("acting %s '%s'...",name.c_str(), buffer+pos-rest));}

  fail=0;
  while(rest>0){
    sent=write(fd,buffer+pos-rest,rest);

    // error?
    if(sent<=0){
      sent=0;
      if(errno!=EAGAIN && errno!=EINTR){
        /*LOG(ALOG_WARN,("Write error in action %s (%s)",name,strerror(errno)));*/
        error=errno;
        fail=1; /* mark an error */
        break;
      }

      if(--retries<0)
        break;

      // sleep and try again
      sl=sleep_time;
      select(0,NULL,NULL,NULL,&sl);
    }
    rest-=sent;
  }
  pos=0;

  // send final \n
  // !!!!!!!!!!!!!!!!!!!!!!!! NO CHECK !!!!!!!!!!!!!!!!!!!!!
  if(rest==0){
    write(fd,"\n",1);
  }
  // not timed out/ no errors?
  if(retries>=0 && !fail)
    return 0;

  if(fail){
    LOG(ALOG_WARN,("Cannot act %s (%s). Disable for %d seconds.",
        name.c_str(),strerror(error),reinit_interval));
  }
  else{
    // timed out
    LOG(ALOG_WARN,("Cannot act %s %d times. Disable for %d seconds.",
        name.c_str(),send_retr,reinit_interval));
  }
  finalize();
  if(Ant_action::global_exe[cmd]!=NULL){
      if(*Ant_action::global_exe[cmd]){
        close(*Ant_action::global_exe[cmd]);
        delete Ant_action::global_exe[cmd];
        Ant_action::global_exe.erase(cmd);
      }
  }
  if(Ant_action::global_err[cmd]!=NULL){
      if(*Ant_action::global_err[cmd]){
        close(*Ant_action::global_err[cmd]);
        delete Ant_action::global_err[cmd];
        Ant_action::global_err.erase(cmd);
      }
  }
  return 1;
}

// com_line - exe [arg1] [arg2] ... \0
//
//  Runs exe.
//
void Ant_action::init(char *com_line, const char *templ){
  // do NOT CARE ABOUT ALREADY INITIALIZED!!!!

  buffer=new char[ANT_ACT_BUF_CHUNK];
  size=ANT_ACT_BUF_CHUNK;
  pos=0;

  // remeber argument template
  argument=templ;


  // remeber command line
  cmd=com_line;

  reinit();
}

void Ant_action::reinit(void){
  //char *const *pnt;

  int i;
  //int pipe_fd[2];
  int *runned_fd;
  char *c; //,*last_zero,*tmp;
  //struct timeval tv;
  char *params[256];
  char cmd_copy[256];


  // check if this exe is runned already
  runned_fd=Ant_action::global_exe[cmd];

  // runned?
  if(runned_fd!=NULL){
    fd=*runned_fd;
    active=1;
    error=0;
    pos=0;
    return;
  }


  strcpy(cmd_copy,cmd.c_str());
  c=cmd_copy;

  // create parameter list
  for(i=0;i<256;++i){

    //skip spaces
    while(*c && isspace(*c))
      ++c;

    //check for eol
    if(*c=='\0')
      break;

    // copy parameter
    params[i]=c;
    while(*c && !isspace(*c))
      ++c;

    //check for eol
    if(*c=='\0')
      break;

    // mark the end
    *c='\0';
    ++c;
  }

  ++i;

  if(i==256){
    LOG(ALOG_WARN,("Warning! paramete list is longer than 256 elements. Shorten it."));
    {/*log("del 02"); */delete params[255];}
    i=255;
  }

  params[i]=NULL;

  string path=actions_path+":/bin:/usr/bin";
  setenv("PATH",path.c_str(),1);
  // execute action module
  fd=ant_exe(params[0], params, NULL);

  if(fd==0){
    LOG(ALOG_WARN,("Cannot execute %s (%s)",name.c_str(),strerror(errno)));
    error=errno;
    active=0;
    return;
  }

  runned_fd=new int(fd);
  Ant_action::global_exe[cmd]=runned_fd;
  active=1;
  error=0;
  pos=0;
} /* reinit */

void Ant_action::finalize(){
  if(buffer)
    {delete[] buffer;}
  buffer=NULL;
  /*  if(cmd)
    {delete[] cmd;}
  */
  if(fd)
    close(fd);

  error=0;
  active=0;
  fd=0;
  size=pos=0;

  time_disabled=time(NULL);

}

Ant_action::~Ant_action(){
  finalize();
/*  if(name)
    {delete name;}
  if(cmd)
    {delete cmd;}
  if(argument)
    {delete argument;}
*/
}

Ant_action::Ant_action(const char *n){
  buffer=NULL;
  error=0;
  active=0;
  fd=0;
  size=pos=0;

  name=n;
}
