/***************************************************************************
 *
 * This is a part of AntMon project
 *
 *            ant_config.cpp
 *
 *  Tue Aug 22 14:33:28 2006
 *  Copyright  2006  Sergey Zhumatiy
 *  serg@parallel.ru
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include <vector>
#include <list>
#include <string>

#include "ant_log.hpp"
#include "ant_mon.hpp"
#include "ant_exe.hpp"

void gentle_exit(int code);

using namespace std;

#define MAX_EXE_SENSOR_NUMBER 128
#define MAX_BUF 1024
#define ANT_INIT_HEADS_TIMEOUT 30

char buffer[MAX_BUF];

void conf_addr(char *c);
void conf_addr_add(char *c);
void conf_name(char *c);
void conf_oname(char *c);
void conf_int(char *c);
void conf_tmout(char *c);
void conf_min(char *c);
void conf_max(char *c);
void conf_minret(char *c);
void conf_maxret(char *c);
void conf_bcount(char *c);
void conf_rcount(char *c);
void conf_ontmout(char *c);
void conf_fail(char *c);
void conf_ret(char *c);
void conf_bad(char *c);
void conf_good(char *c);
void conf_sbad(char *c);
void conf_sgood(char *c);
void conf_dep(char *c);
void conf_var(char *c);
void conf_action(char *c);
void conf_act_line(char *c);
void conf_addrs(char *c);
void conf_sensor(char *c);
void conf_aset(char *c);
void conf_nset(char *c);
void conf_def(char *c);
void conf_loglevel(char *c);
void conf_recon_int(char *c);
void conf_recon_tmout(char *c);
void conf_conn(char *c);
void conf_conn_err(char *c);
void conf_port(char *c);
void conf_hport(char *c);
void conf_heads(char *c);
void conf_topo(char *c);
void conf_head_death(char *c);
void conf_any_head_death(char *c);
void conf_head_tmout(char *c);
void conf_head_ping(char *c);
void conf_mons(char *c);
void conf_any(char *c);

void new_sensor();
void set_actions(list < Ant_action * >*, char *);

// vector of lists of sensors (sorted by ip)
vector < vector < class Ant_mon_sensor * >*>Sensors_by_ip;
// vector of connection (sorted by ip), accorded with Sensors_by_ip
vector < class Ant_connection * >Conn_by_ip;

list < Ant_action * >on_conn;
list < Ant_action * >on_conn_err;

map < string, class Ant_action * >Actions;
map < string, string > Variables;
list < string > Mass_addr;

//#define close(x) {fprintf(stderr,"!!close(%d)\n",x);close(x);}
/*
// head servers, which are needed to be pinged
short int     neighbors[ANT_MAX_HEADS];
*/

int head_port = ANT_HEAD_PORT;
int agent_port = ANT_PORT;

extern time_t global_time;

// dead head servers
short int dead_heads[ANT_MAX_HEADS];
// heads, which are pinging me
short int head_ins[ANT_MAX_HEADS];
// heads, which I must to ping
short int head_outs[ANT_MAX_HEADS];

// head server ping timeout
long unsigned int head_tmout;
// head server ping interval
long unsigned int head_ping;


// actions on near head server death
list < Ant_action * >head_death;
// actions on any head server death
list < Ant_action * >head_death_any;

// connections to head servers
vector < class Ant_head * >heads;

// my index in head server list
int my_head_index;
string hostname, domainname;

string current_act_line = "";

struct sensor_templ {
    string addr;
    string name;
    string out_name;
    string last_name;
    long int interval;
    long int tmout;
    long int tmout_interval;
    float min;
    float max;
    float min_ret;
    float max_ret;
    int bcount;
    int rcount;
    int passive;
    short int heads[ANT_MAX_HEADS];
    string on_fail;
    string on_ret;
    string on_tmout;
    string on_dep;
    string on_bad;
    string on_good;
    string on_any;
    string on_single_bad;
    string on_single_good;
    string dep;
};

struct sensor_templ cur;

extern long int recon_interval;
extern long int recon_timeout;
//
// Reads a line (end with EOF or EOL) no longer than len
// Only syncronous fd!!!
//
// Returns 0 on success
//
int readln(int fd, int len, char *to)
{
    int status;
    int was_readed=0;

    while ((status = read(fd, to, 1)) > 0) {

        was_readed=1;
        // buffer overrun?
        if (--len < 1) {
            *to = '\0';
            return 1;
        }

        // eol?
        if (*to == '\n') {
            *to = '\0';
            return 0;
        }
        ++to;
    }
    // eof?
    if (status == 0) {
        if (len > 1)
            ++to;
        *to = '\0';
        if(was_readed)
            return 0;

        // fatal eof
        return 1;
    }

    // error...
    *to = '\0';
    return 1;
}

struct conf_p {
    const char *name;
    void (*handler) (char *);
};

struct conf_p conf_params[] = {

    // per-parameter atributes
    {"addr", conf_addr},        /* set address for next 'names' or 'name_set' */
    {"addr_add", conf_addr_add},/* ADDS addresses for list in 'addressess' */
    {"name", conf_name},        /* set name for next 'addresses' */
    {"addresses", conf_addrs},  /* set list of addresses for next 'names' */
                                /* RESET all previous addresses! (see addr_add) */
    {"names", conf_sensor},     /* !! 'sensors' alias (compatibility ONLY) */
    {"sensor", conf_sensor},    /* creates sensors with same profile on all nodes,
                                 * saved in 'addresses' */
    {"addr_set", conf_aset},    /* create several sensors with THE SAME NAME (space is delimiter) */
    {"name_set", conf_nset},    /* create several sensors on ONE ADDRESS(see addr) (; is delimiter) */
    {"out_name", conf_oname},   /* SENS output sensor name */
    {"interval", conf_int},     /* SENS interval to check sensor */
    {"tmout", conf_tmout},      /* SENS sensor timeout */
    {"min", conf_min},          /* SENS minimum normal */
    {"max", conf_max},          /* SENS maximum normal */
    {"min_ret", conf_minret},   /* SENS value to try to return from minimum */
    {"max_ret", conf_maxret},   /* SENS value to try to return from maximum */
    {"ret_count", conf_rcount}, /* SENS number of good values to return to normal */
    {"bad_count", conf_bcount}, /* SENS number of bad values to be fail */

    {"on_fail", conf_fail},     /* ACTION on sensor fail */
    {"on_ret", conf_ret},       /* ACTION on return to normal */
    {"on_tmout", conf_ontmout}, /* ACTION on sensor timeout */
    {"on_bad", conf_bad},       /* ACTION on sensor bad value */
    {"on_good", conf_good},     /* ACTION on sensor good value */
    {"on_any", conf_any},       /* ACTION on any sensor value */
    {"on_single_bad", conf_sbad},       /* ACTION on single sensor bad value */
    {"on_single_good", conf_sgood},     /* ACTION on single sensor good value */
    {"on_conn", conf_conn},     /* ACTION on connection */
    {"on_conn_err", conf_conn_err},     /* ACTION on connection error */
    {"depend_on", conf_dep},    /* dependings for this sensor */

    //
    // global attributes and definitions
    //
    {"defaults", conf_def},     /* reset all settings */
    {"action", conf_action},    /* action definition (name) */
    {"act_line", conf_act_line},        /* action line definition */
    {"recon_interval", conf_recon_int}, /* agents reconnection interval */
    {"recon_tmout", conf_recon_tmout},  /* agents reconnection timeout */
    {"loglevel", conf_loglevel},        /* log level */
    {"var", conf_var},          /* custom variable */
    {"port", conf_port},        /* agent port */

    // heads...
    {"head_port", conf_hport},  /* head servers port */
    {"heads", conf_heads},      /* list of head servers */
    {"topo", conf_topo},        /* monitoring topology (route of pings - indexes from 1) */
    {"head_tmout", conf_head_tmout},    /* timeout for other heads */
    {"head_ping", conf_head_ping},      /*  interval to ping other heads */
    {"mons", conf_mons},        /* SENSOR list of serving head servers (indexes from 1) */

    {"on_head_death", conf_head_death}, /* ACTION on head server, which pings me, death */
    {"on_any_head_death", conf_any_head_death}, /* ACTION on any head server death */

    {NULL, conf_def}
};

//
//  Try to load config file
//
//  Returns 0 on success
//          1 on error
//
int load_config(const char *path)
{

    int fd = open(path, O_RDONLY);
    int i, j, start;
    char tmp[1024];

    if (fd < 0) {
        return 1;
    }

    // reset defaults
    conf_def(NULL);

    gethostname(tmp, 1023);
    hostname = tmp;
    getdomainname(tmp, 1023);
    domainname = tmp;

    my_head_index = -1;

/*
  neighbors[0]=-1;
*/
    dead_heads[0] = -1;
    head_ins[0] = -1;
    head_outs[0] = -1;


    // loop on every line
    while (!readln(fd, MAX_BUF, buffer)) {

        i = 0;
        // read leading spaces
        while (buffer[i] && isspace(buffer[i]))
            ++i;

        start = i;

        // skip comment or empty line
        if (buffer[i] == '#' || buffer[i] == '\0')
            continue;

        // read instruction
        while (buffer[i] && !isspace(buffer[i]))
            ++i;

        buffer[i] = '\0';

        // recognise intruction
        for (j = 0; conf_params[j].name != NULL; ++j) {

            // does instruction match?
            if (!strcmp(buffer + start, conf_params[j].name)) {
                conf_params[j].handler(buffer + i + 1);
                break;
            }
        }

        // was instruction recognized?
        if (conf_params[j].name != NULL)
            continue;

        // restore space symbol and log error
        buffer[i] = ' ';
        LOG(ALOG_WARN, ("Bad string in config: '%s'. Ignored", buffer));
    }
    //eof? (tricky check)...
    if (*buffer == '\0')
        return 0;

    LOG(ALOG_WARN, ("Error loading config '%s': %s", path, strerror(errno)));
    return 1;
}

//
//  hook for 'addr' instruction
//
void conf_addr(char *c)
{

    // just copy address
    cur.addr = c;
}

//
//  hook for 'name' instruction
//
void conf_name(char *c)
{

    // just copy name
    cur.name = c;
    cur.out_name = c;
}

//
//  hook for 'out_name' instruction
//
void conf_oname(char *c)
{

    // just copy out_name
    cur.out_name = c;
}

//
//  hook for 'interval' instruction
//
void conf_int(char *c)
{

    // copy interval
    if (sscanf(c, "%ld", &cur.interval) < 1) {

        // bad interval
        LOG(ALOG_WARN, ("Warning! Bad config line '%s'. Ignored.", buffer));
    }
}

//
//  hook for 'tmout' instruction
//
void conf_tmout(char *c)
{

    // copy timeout
    if (sscanf(c, "%ld", &cur.tmout) < 1) {

        // bad timeout
        LOG(ALOG_WARN, ("Warning! Bad config line '%s'. Ignored.", buffer));
    }
}

//
//  hook for 'min' instruction
//
void conf_min(char *c)
{

    // copy min
    if (sscanf(c, "%f", &cur.min) < 1) {

        // bad min
        LOG(ALOG_WARN, ("Warning! Bad config line '%s'. Ignored.", buffer));
    }
    cur.min_ret = cur.min;
}

//
//  hook for 'max' instruction
//
void conf_max(char *c)
{

    // copy max
    if (sscanf(c, "%f", &cur.max) < 1) {

        // bad max
        LOG(ALOG_WARN, ("Warning! Bad config line '%s'. Ignored.", buffer));
    }
    cur.max_ret = cur.max;
}

//
//  hook for 'min_ret' instruction
//
void conf_minret(char *c)
{

    // copy min_ret
    if (sscanf(c, "%f", &cur.min_ret) < 1) {

        // bad min_ret
        LOG(ALOG_WARN, ("Warning! Bad config line '%s'. Ignored.", buffer));
    }
}

//
//  hook for 'max_ret' instruction
//
void conf_maxret(char *c)
{

    // copy max_ret
    if (sscanf(c, "%f", &cur.max_ret) < 1) {

        // bad max_ret
        LOG(ALOG_WARN, ("Warning! Bad config line '%s'. Ignored.", buffer));
    }
}

//
//  hook for 'bad_count' instruction
//
void conf_bcount(char *c)
{

    // copy bad_count
    if (sscanf(c, "%d", &cur.bcount) < 1) {

        // bad bad_count
        LOG(ALOG_WARN, ("Warning! Bad config line '%s'. Ignored.", buffer));
    }
}

//
//  hook for 'port' instruction
//
void conf_port(char *c)
{

    // copy port
    if (sscanf(c, "%d", &agent_port) < 1) {

        LOG(ALOG_WARN, ("Warning! Bad config line '%s'. Ignored.", buffer));
    }
}

//
//  hook for 'head_port' instruction
//
void conf_hport(char *c)
{

    // copy port
    if (sscanf(c, "%d", &head_port) < 1) {

        LOG(ALOG_WARN, ("Warning! Bad config line '%s'. Ignored.", buffer));
    }
}

//
//  hook for 'ret_count' instruction
//
void conf_rcount(char *c)
{

    // copy ret_count
    if (sscanf(c, "%d", &cur.rcount) < 1) {

        // bad ret_count
        LOG(ALOG_WARN, ("Warning! Bad config line '%s'. Ignored.", buffer));
    }
}

//
//  hook for 'on_fail' instruction
//
void conf_fail(char *c)
{

    // just copy action names
    cur.on_fail = c;

}

//
//  hook for 'on_ret' instruction
//
void conf_ret(char *c)
{

    // just copy action names
    cur.on_ret = c;

}

//
//  hook for 'on_any' instruction
//
void conf_any(char *c)
{

    // just copy action names
    cur.on_any = c;

}

//
//  hook for 'on_ontmout' instruction
//
void conf_ontmout(char *c)
{

    // just copy action names
    cur.on_tmout = c;

}

//
//  hook for 'on_bad' instruction
//
void conf_bad(char *c)
{

    // just copy action names
    cur.on_bad = c;

}

//
//  hook for 'on_good' instruction
//
void conf_good(char *c)
{

    // just copy action names
    cur.on_good = c;

}

//
//  hook for 'on_single_bad' instruction
//
void conf_sbad(char *c)
{

    // just copy action names
    cur.on_single_bad = c;

}

//
//  hook for 'on_single_good' instruction
//
void conf_sgood(char *c)
{

    // just copy action names
    cur.on_single_good = c;

}

//
//  hook for 'depend_on' instruction
//
void conf_dep(char *c)
{

    // just copy dependences names
    cur.dep = c;

}

//
//  hook for 'var' instruction
//
void conf_var(char *c)
{
    char *name;

    // skip spaces
    while (*c && isspace(*c))
        ++c;

    if (*c == '\0') {
        LOG(ALOG_WARN,
            ("Warning! Bad variable declaration: '%s'. Ignore.", c));
        return;
    }

    name = c;

    // read name
    while (*c && !isspace(*c))
        ++c;

    // not empty value?
    if (*c != '\0') {
        *c = '\0';
        ++c;
    }

    // skip leading spaces
    while (*c && isspace(*c))
        ++c;

    Variables[name] = string(c);
}



//
//  hook for 'act_line' instruction
//
// remembers template line for action(s)
void conf_act_line(char *c)
{

    current_act_line = c;
}

//
//  hook for 'action' instruction
//
// adds new action to Actions hash
void conf_action(char *c)
{

    char *exe;

    Ant_action *act, *tmp;

    // read action name
    for (exe = c; *exe && !isspace(*exe);)
        ++exe;

    // check format
    if (*exe == '\0') {
        LOG(ALOG_WARN,
            ("Warning! Bad action format (must be: action name exe [args] (%s). Ignored",
             buffer));
        return;
    }
    *exe = '\0';

    // create new action (c=name)
    act = new Ant_action(c);

    // skip spaces
    ++exe;
    while (*exe && isspace(*exe))
        ++exe;


    tmp = Actions[string(c)];
    if (tmp != NULL) {
        LOG(ALOG_WARN,
            ("Warning! Action '%s' is defined already. Ignore.", c));
        {                       /*log("del a8"); */
            delete act;
        }
        return;
    }

    //Actions.erase(string(c));
    act->init(exe, current_act_line.c_str());
    Actions[string(c)] = act;

    LOG(ALOG_CONF, ("Action '%s' defined.", c));
    //  p->on_fail.push_back(act);
}


//
//  hook for 'interval_tmout' instruction
//
//  set inteval for checking timed out resources
void conf_int_tmout(char *c)
{

    if (sscanf(c, "%ld", &cur.tmout_interval) < 1)
        LOG(ALOG_CONF, ("Invalid interval, must be a number: %s", c));

    if (cur.tmout_interval < 0)
        LOG(ALOG_CONF, ("Invalid interval, must be positive: %s", c));
}

//
//  hook for 'recon_interval' instruction
//
//  set inteval for reconnecting to agents
void conf_recon_int(char *c)
{

    if (sscanf(c, "%ld", &recon_interval) < 1)
        LOG(ALOG_CONF,
            ("Invalid reconection interval, must be a number: %s", c));

    if (recon_interval < 0)
        LOG(ALOG_CONF,
            ("Invalid reconection interval, must be positive: %s", c));
}

//
//  hook for 'reconn_tmout' instruction
//
//  set inteval for reconnecting to agents
void conf_recon_tmout(char *c)
{

    if (sscanf(c, "%ld", &recon_timeout) < 1)
        LOG(ALOG_CONF,
            ("Invalid reconnection timeout, must be a number: %s", c));

    if (recon_timeout < 0)
        LOG(ALOG_CONF,
            ("Invalid reconnection timeout, must be positive: %s", c));
}

//
//  hook for 'on_conn' instruction
//
//  set GLOBAL actions for connection establish
void conf_conn(char *c)
{

    set_actions(&(on_conn), c);
}

//
//  hook for 'on_conn_err' instruction
//
//  set GLOBAL actions for connection fail
void conf_conn_err(char *c)
{

    set_actions(&(on_conn_err), c);
}

//
//  hook for 'log_level' instruction
//
//  set log level
void conf_loglevel(char *c)
{

    if (sscanf(c, "%d", &log_level) < 1)
        LOG(ALOG_CONF, ("Invalid loglevel, must be a number: %s", c));
}




//
//  hook for 'defaults' instruction
//
//  reset settings
//
void conf_def(char *c)
{

    cur.addr = "";
    cur.name = "";
    cur.out_name = "";
    cur.last_name = "";
    cur.interval = 600;
    cur.tmout = 30;
    cur.tmout_interval = 600;
    cur.min = 0.0;
    cur.max = 1.0;
    cur.min_ret = 0.0;
    cur.max_ret = 1.0;
    cur.bcount = 1;
    cur.rcount = 1;
    cur.passive = 1;
    cur.on_fail = "";
    cur.on_ret = "";
    cur.on_tmout = "";
    cur.on_bad = "";
    cur.on_good = "";
    cur.on_any = "";
    cur.on_single_bad = "";
    cur.on_single_good = "";
    cur.dep = "";
    cur.heads[0] = 0;
    cur.heads[1] = -1;
}

//
//  hook for 'addresses' instruction
//
//  just remember addresses for later call of 'names'
//
void conf_addrs(char *c)
{
    Mass_addr.clear();

    conf_addr_add(c);

}

//
//  hook for 'addr_add' instruction
//
//  remember new addresses for later call of 'names'
//
void conf_addr_add(char *c)
{
    char *p1;
    for (;;) {

        // skip spaces
        while (*c && isspace(*c))
            ++c;

        p1 = c;
        while (*c && !isspace(*c))
            ++c;

        // empty addr?
        if (p1 == c)
            return;

        // last sensor?
        if (*c == '\0') {
            // remember address
            Mass_addr.push_back(string(p1));
            return;
        }

        *c = '\0';

        // remember address
        Mass_addr.push_back(string(p1));

        ++c;
    }
}

//
//  hook for 'heads' instruction
//
//  Sets list of head servers
//
void conf_heads(char *c)
{
    char *name, *p1;
    int num = -1;
    string str;
    Ant_head *h;

    for (;;) {

        // skip spaces
        while (*c && isspace(*c))
            ++c;

        p1 = c;
        while (*c && !isspace(*c) && (*c != ','))
            ++c;

        // empty server name?
        if (p1 == c)
            return;

        // p1=server  name
        name = p1;

        // last server name? Just mark it...
        if (*c == '\0')
            p1 = NULL;

        // mark server name end.
        *c = '\0';

        // create new head server!
        str = name;

        // increase index
        ++num;

        // push its name into list
        //head_names.push_back(str);

        // create connection for it
        h = new Ant_head;
        // copy name
        h->name = str;
        h->out_conn = NULL;     //new Ant_connection;
        //h->out_conn->port=head_port;
        h->in_conn = new Ant_connection;
        h->last_ping = h->next_ping = global_time + ANT_INIT_HEADS_TIMEOUT;
        h->state = Head_No_Init;
        h->timed_out = 0;
        //h->out_conn->init(str.c_str());
        h->in_conn->init(str.c_str());

        heads.push_back(h);

/*
    Ant_connection *conn=new Ant_connection;
    conn->name=new char[str.length()+1];
    strcpy(conn->name,str.c_str());
    head_conn.push_back(conn);
*/

        // check if it is me!
        if (str == hostname || str == domainname)
            if (my_head_index != -1) {
                LOG(ALOG_CONF,
                    ("My name is listed more than one time in 'heads'. Exiting.\n"));
                gentle_exit(1);
            }
            else {
                LOG(ALOG_DEBUG2, (">> My index: %d\n", num+1));
                my_head_index = num;
            }

/*
    // add entry for timeout
    tm=new int;
    *tm=global_time;
    head_last_ping.push_back(tm);
*/

        // it was last sensor?
        if (p1 == NULL)
            return;

        ++c;
    }
}                               // conf_heads

//
//  hook for 'head_tmout' instruction
//
void conf_head_tmout(char *c)
{

    // copy timeout
    if (sscanf(c, "%lu", &head_tmout) < 1) {

        // bad timeout
        LOG(ALOG_WARN, ("Warning! Bad config line '%s'. Ignored.", buffer));
    }
}                               // conf_head_tmout

//
//  hook for 'head_ping' instruction
//
void conf_head_ping(char *c)
{

    // copy timeout
    if (sscanf(c, "%lu", &head_ping) < 1) {

        // bad timeout
        LOG(ALOG_WARN, ("Warning! Bad config line '%s'. Ignored.", buffer));
    }
}                               // conf_head_ping


//
//  hook for 'topo' instruction
//
//  Sets the head servers topology
//
void conf_topo(char *c)
{
    char *name, *m1, old;
    int num, prev = -1, ins = 0, outs = 0, notlast;

    notlast = 1;
    name = c;
    while (notlast) {

        // skip spaces
        while (*c && isspace(*c))
            ++c;

        m1 = c;
        while (*c && !isspace(*c) && (*c != ','))
            ++c;

        // empty head number?
        if (m1 == c)
            return;

        // last server name? Just mark it...
        if (*c == '\0')
            notlast = 0;

        // mark server name end.
        old = *c;
        *c = '\0';

        // get the number
        num = 0;
        while (*m1) {
            if (*m1 < '0' || *m1 > '9') {
                LOG(ALOG_CONF,
                    ("Bad head server number in 'topo' instruction (must be a number): '%s'\n",
                     name));
                return;
            }
            num *= 10;
            num += *m1 - '0';
            ++m1;
        }

        --num;                  // count from1 - > from 0

        //    LOG(ALOG_DEBUG2,(">> TOPO %d\n",num));

        // is it me?
        if ((num == my_head_index) && (prev != -1)) {

            if (num >= ANT_MAX_HEADS) {
                LOG(ALOG_CONF,
                    ("Too many head servers, pinging me in 'topo' instruction (%d max)\n",
                     ANT_MAX_HEADS));
                return;
            }

            head_ins[ins] = prev;

            //      LOG(ALOG_DEBUG2,(">> ins[%d]= = %d\n",ins,prev));

            // mark the end
            if (++ins < ANT_MAX_HEADS)
                head_ins[ins] = -1;
        }
        // must I ping it?
        else if (prev == my_head_index) {
            if (outs >= ANT_MAX_HEADS) {
                LOG(ALOG_CONF,
                    ("Too many head servers, I must to ping in 'topo' instruction (%d max)\n",
                     ANT_MAX_HEADS));
                return;
            }
            head_outs[outs] = num;

            //      LOG(ALOG_DEBUG2,(">> outs[%d] = %d\n",outs,num));

            // mark the end
            if (++outs < ANT_MAX_HEADS)
                head_outs[outs] = -1;
        }

        prev = num;
        ++c;
    }
}

//
//  hook for 'on_head_death' instruction
//
//  sets action on head server death
//
void conf_head_death(char *c)
{
    set_actions(&head_death, c);
}

//
//  hook for 'on_any_head_death' instruction
//
//  sets action on head server death
//
void conf_any_head_death(char *c)
{
    set_actions(&head_death_any, c);
}

//
//  hook for 'mons' instruction
//
//  points list of responding servers numbers on this sensor
//
void conf_mons(char *c)
{
    char *name, *m1;
    int num, i = 0, notlast;

    notlast = 1;
    name = c;
    while (notlast) {

        // skip spaces
        while (*c && isspace(*c))
            ++c;

        m1 = c;
        while (*c && !isspace(*c) && (*c != ','))
            ++c;

        // empty head number?
        if (m1 == c)
            return;


        // last server name? Just mark it...
        if (*c == '\0')
            notlast = 0;

        // mark server name end.
        *c = '\0';

        // get the number
        num = 0;
        while (*m1) {
            if (*m1 < '0' || *m1 > '9') {
                LOG(ALOG_CONF,
                    ("Bad head server number in 'mons' instruction (must be a number): '%s'\n",
                     name));
                return;
            }
            num *= 10;
            num += *m1 - '0';
            ++m1;
        }
        --num;                  // numbering from 1 -> from 0

        if(num>heads.size()){
            LOG(ALOG_CONF,
                ("Bad head index: %d. Only %d heads are declared. Ignore.\n",
                num,heads.size()
                )
            );
        }
        else{
            // add a number to sensors head servers list
            if (i >= ANT_MAX_HEADS) {
                LOG(ALOG_CONF,
                    ("Too many head servers in 'mons' instruction (%d max)\n",
                     ANT_MAX_HEADS));
                return;
            }
            cur.heads[i] = num;
            // mark the end
            if (++i < ANT_MAX_HEADS)
                cur.heads[i] = -1;
        }
        ++c;
    }
}

//
//  hook for 'sensor' instruction
//
//  create several new sensors with different addresses and same names
//  instruction 'addr' MUST be called before
//
void conf_sensor(char *c)
{
    char *p1;
    list < string >::iterator it, end;

    end = Mass_addr.end();
    for (;;) {

        // skip spaces
        while ((*c != '\0') && isspace(*c))
            ++c;

        p1 = c;
        while ((*c != '\0') && (*c != ';'))
            ++c;

        // empty name?
        if (p1 == c)
            return;

        // p1 = full sensor name
        cur.name = p1;

        // last name? Just mark it...
        if (*c == '\0')
            p1 = NULL;

        // mark sebsor name end.
        *c = '\0';

        // create new sensor!
        for (it = Mass_addr.begin(); it != end; ++it) {

            // recall address...
            cur.addr = *it;
            new_sensor();
        }

        // it was last sensor?
        if (p1 == NULL)
            return;

        ++c;
    }
}                               /* conf_names */

//
//  hook for 'addr_set' instruction
//
//  create several new sensors with different addresses
//
void conf_aset(char *c)
{
    char *p1;

    for (;;) {

        // skip spaces
        while (*c && isspace(*c))
            ++c;

        p1 = c;
        while (*c && !isspace(*c))
            ++c;

        // empty addr?
        if (p1 == c)
            return;

        // last sensor?
        if (*c == '\0') {
            // create new sensor!
            cur.addr = p1;
            new_sensor();
            return;
        }

        *c = '\0';

        // create new sensor!
        cur.addr = p1;
        new_sensor();

        ++c;
    }
}                               /* conf_aset */

//
//  hook for 'name_set' instruction
//
//  create several new sensors with different names
//
void conf_nset(char *c)
{

    char *p1;

    for (;;) {

        // skip spaces
        while (*c && isspace(*c))
            ++c;

        p1 = c;
        while (*c && (*c != ';'))
            ++c;

        // empty name?
        if (p1 == c)
            return;

        cur.name = p1;
        if (*c == '\0')
            p1 = NULL;

        *c = '\0';

        new_sensor();

        if (p1 == NULL)
            return;

        ++c;
    }
}                               /* conf_nset */

//
//  create new sensor from data, contained in cur
//
void new_sensor()
{
    int i;
    unsigned int max;
    class Ant_mon_sensor *p = new class Ant_mon_sensor;
    vector < class Ant_mon_sensor * >*v;
    char loc_buffer[MAX_BUF];
    const char *cp, *last;
    string tmp_str;

    p = new Ant_mon_sensor;

    LOG(ALOG_DEBUG, ("Defining sensor '%s' on '%s'.",
                     cur.name.c_str(), cur.addr.c_str()));
    // copy from cur
    p->name=cur.name.c_str();
    p->out_name=cur.out_name.c_str();

    for (last = cp = p->name.c_str(); *cp != '\0'; ++cp)
        if (isspace(*cp))
            last = cp + 1;
    p->last_name=last;

    // set initial value
    p->value=0.1234;

    if (cur.min > cur.max) {
        LOG(ALOG_WARN,
            ("Warning! Minimum value of '%s' is greater than maximum. Reset minimum.",
             p->name.c_str()));
        cur.min = cur.max;
    }

    p->interval = cur.interval;
    p->timeout = cur.tmout;
    p->tmout_interval = cur.tmout;
    p->min_value = cur.min;
    p->max_value = cur.max;
    p->ret_min_value = cur.min_ret;
    p->ret_max_value = cur.max_ret;
    p->passive = cur.passive;
    p->ret_count = cur.rcount;
    p->bad_count = cur.bcount;
    for (i = 0; i < ANT_MAX_HEADS; ++i)
        p->heads[i] = cur.heads[i];

    // set initial values
    p->id = -1;
    p->data_sent = 0;
    p->value = 0.0;

    // actions...
    tmp_str=cur.on_any+" "+cur.on_ret;
    strncpy(loc_buffer, tmp_str.c_str(), MAX_BUF - 1);
    loc_buffer[MAX_BUF - 1] = '\0';
    set_actions(&(p->on_ret), loc_buffer);

    tmp_str=cur.on_any+" "+cur.on_fail;
    strncpy(loc_buffer, tmp_str.c_str(), 1023);
    loc_buffer[MAX_BUF - 1] = '\0';
    set_actions(&(p->on_fail), loc_buffer);

    tmp_str=cur.on_any+" "+cur.on_bad;
    strncpy(loc_buffer, tmp_str.c_str(), 1023);
    loc_buffer[MAX_BUF - 1] = '\0';
    set_actions(&(p->on_bad), loc_buffer);

    tmp_str=cur.on_any+" "+cur.on_good;
    strncpy(loc_buffer, tmp_str.c_str(), 1023);
    loc_buffer[MAX_BUF - 1] = '\0';
    set_actions(&(p->on_good), loc_buffer);

    tmp_str=cur.on_any+" "+cur.on_single_bad;
    strncpy(loc_buffer, tmp_str.c_str(), 1023);
    loc_buffer[MAX_BUF - 1] = '\0';
    set_actions(&(p->on_single_bad), loc_buffer);

    tmp_str=cur.on_any+" "+cur.on_single_good;
    strncpy(loc_buffer, tmp_str.c_str(), 1023);
    loc_buffer[MAX_BUF - 1] = '\0';
    set_actions(&(p->on_single_good), loc_buffer);
    strncpy(loc_buffer, cur.on_tmout.c_str(), 1023);
    loc_buffer[MAX_BUF - 1] = '\0';
    set_actions(&(p->on_tmout), loc_buffer);
    strncpy(loc_buffer, cur.on_dep.c_str(), 1023);
    loc_buffer[MAX_BUF - 1] = '\0';
    set_actions(&(p->on_dep), loc_buffer);


    max = Conn_by_ip.size();
    for (i = 0; i < (int) max; ++i) {

        // is there such connection already?
        if (cur.addr == Conn_by_ip[i]->name) {

            // use existent
            Sensors_by_ip[i]->push_back(p);
            p->conn = Conn_by_ip[i];
            return;
        }
    }

    // no corresponding connection yet... Create it!
    Ant_connection *conn = new Ant_connection;
    conn->port = agent_port;

    conn->name = cur.addr;

    //  LOG(ALOG_DEBUG2,("My head index=%d",my_head_index));
    // check if sensor is controlled by me by default
    if (cur.heads[0] == my_head_index) {
        p->is_own = 1;
        p->controlled = 'y';
        //    LOG(ALOG_DEBUG2,("CONTROLLED"));
    }
    else {
        p->is_own = 0;
        p->controlled = 'n';
        //    LOG(ALOG_DEBUG2,("*NOT* CONTROLLED"));
    }

    //connection.
    Conn_by_ip.push_back(conn);
    v = new vector < class Ant_mon_sensor * >;
    Sensors_by_ip.push_back(v);

    // add sensor to lists...
    v->push_back(p);
    p->conn = conn;

}

//
//  Writes actions poinetrs to Ant_mon_sensor
//  Reads action name s from second argument
//
void set_actions(list < Ant_action * >*actlist, char *c)
{
    char *p;
    Ant_action *act;

    if (*c == '\0')
        return;

    LOG(ALOG_CONF, ("Actions: %s", c));

    // for every action
    for (;;) {

        // skip spaces
        while (*c && isspace(*c))
            ++c;

        // finished processing actions?
        if (*c == '\0')
            return;

        // read action name
        p = c;
        while (*c && !isspace(*c))
            ++c;

        // last action name?
        if (*c != '\0') {
            *c = '\0';
            ++c;
        }

        LOG(ALOG_CONF, ("Action: %s", p));

        act = Actions[p];
        if (act == NULL) {
            Actions.erase(p);
            LOG(ALOG_WARN, ("Action '%s' is not defined. Skip.", p));
            continue;
        }
        actlist->push_back(act);
    }
}
