/***************************************************************************
 *
 * This is a part of AntMon project
 *
 *            ant_mon.cpp
 *
 *  Tue Aug 22 14:33:28 2006
 *  Copyright  2006  Sergey Zhumatiy
 *  serg@parallel.ru
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define _XOPEN_SOURCE 600
#include <unistd.h>
#include <signal.h>
#include <stdarg.h>
#include <syslog.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <resolv.h>
#include <netdb.h>
#include <time.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include <string>
#include <list>
#include <vector>
#include <queue>
#include <map>

#include "ant_net.hpp"
#include "ant_mon.hpp"
#include "ant_log.hpp"
#include "ant_act.hpp"
/*#include "ant_hashes.hpp"*/
#include "ant_subst.hpp"
#include "ant_time.hpp"
#include "ant_exe.hpp"

#define MAX_SERV 16

#define USE_RLIMIT

//#define close(x) {fprintf(stderr,"!!close(%d)\n",x);close(x);}

using namespace std;

#ifdef USE_RLIMIT
#include <sys/resource.h>
struct rlimit cpulim;
#endif

#ifndef O_LARGEFILE
#define O_LARGEFILE 0
#endif

void sig_log_handler(int);
void sig_exit_handler(int);
void sig_urg_handler(int);
void exe_sig_chld_handler(int);
void sig_usr1_handler(int);
void sig_usr2_handler(int);
void sig_hup_handler(int);

void create_pid(void);
int load_config(const char *);
void create_connections();
int GetNonblockSocket();
//int  GetAddress(struct addrinfo *ptr, char *host, int port);
void process_message(Ant_connection * conn);
void process_timed_out(Ant_mon_sensor * param);
void process_bad_value(Ant_mon_sensor * param);
void process_good_value(Ant_mon_sensor * param);
void process_connected(int i);
void process_conn_error(int i);

void process_connected_head(int i);
void process_conn_error_head(int i, int my_message);
void process_message_head(Ant_connection * conn);

long int new_id();
void daemonize();
void gentle_exit(int code);
#define CHLD_INTERVAL 5

void fill_vars(map < string, string > *vars, Ant_mon_sensor * p);
void fill_vars_conn(map < string, string > *vars, Ant_connection * p);
void fill_vars_head(map < string, string > *vars, Ant_head * p);
void mark_dependent(Ant_connection * conn, Ant_mon_sensor_state state);
void on_dep(Ant_mon_sensor * param);
void on_tmout(Ant_mon_sensor * param);
void on_single_bad(Ant_mon_sensor * param);
void on_single_good(Ant_mon_sensor * param);
void on_ret(Ant_mon_sensor * param);
void on_fail(Ant_mon_sensor * param);
void on_bad(Ant_mon_sensor * param);
void on_good(Ant_mon_sensor * param);
void on_connect(int i);
void on_conn_error(int i);
void on_any_head_death(int i);
void on_head_death(int i);


/*
int multi_accept(int *server_s, int nb_server_s,
                 struct sockaddr *addr=NULL, socklen_t *len=NULL);
int build_server_socket(struct addrinfo *ai);
int get_name_and_port(int server_s, char *shost, char *sport);
*/
void config_reload();

/////////////////////////////////////////////////////////////////
//
//  Global variabbles
//
/////////////////////////////////////////////////////////////////

#define DEBUG
//#define DEBUG //

#define MAX_ID 2147483647
//             = 2**31-1

#define MAX_BUF 10240

//typedef std::vector<class Ant_mon_sensor *> Ant_param_vector;

// vector of lists of parameters (sorted by ip)
extern std::vector < std::vector < class Ant_mon_sensor * >*>Sensors_by_ip;
// vector of connection (sorted by ip), accorded with Sensors_by_ip
extern std::vector < class Ant_connection * >Conn_by_ip;

// list of actions on connections fails/establishes
extern std::list < Ant_action * >on_conn;
extern std::list < Ant_action * >on_conn_err;


// list of sens requests
list < class Ant_mon_sensor * >Waited_params;

// hash of parameters
map < long int, void *>params_by_id;


extern int head_port;
extern int agent_port;

// dead head servers
extern short int dead_heads[ANT_MAX_HEADS];
// heads, which are pinging me
extern short int head_ins[ANT_MAX_HEADS];
// heads, which I must to ping
extern short int head_outs[ANT_MAX_HEADS];

// actions on near head server death
extern list < Ant_action * >head_death;
// actions on any head server death
extern list < Ant_action * >head_death_any;


// head server ping timeout
extern long unsigned int head_tmout;
// head server ping interval
extern long unsigned int head_ping;

// connections to head servers
extern vector < class Ant_head * >heads;

// my index in head server list
extern int my_head_index;


// hash of actions
extern map < string, Ant_action * >Actions;
typedef map < string, Ant_action * >::iterator Act_iter;

extern map < string, string > Variables;

// sleep interval
struct timeval sleep_time;
struct timeval def_sleep_time = { 0, 1000000 };

// flag, if any dependence was changed
int was_change = 0;
// value of maximum interval adding
long int max_interval_change;

// interval to reconnect
long int recon_interval = 60;
// reconnect timeout
long int recon_timeout = 300;

static long int _id = 0;
static char tmpbuf[MAX_BUF];

static char char_buf[128];

/////////////////////////////////////////////////////////////////
//
//  Functions
//
/////////////////////////////////////////////////////////////////
int test_param(Ant_mon_sensor *param, long int id, int bad_val);



//  Main
//
/////////////////////////////////////////////////////////////////

int main(int argc, char *argv[])
{
    int num_of_conn = 0;        // current number of conections
    int i, j, max, flag;        // ....
    int interval_add;           // random reconnection increment
    int daemon = 0;             // daemon flag
    //int sd;                     // listen sockets
    int mon_sd;                 // listen sockets
    long int id;                // current param id
    char serv_name[NI_MAXHOST]; // name of remote head server

    char conf_file[1024] = "/etc/mon.conf";

    //struct addrinfo hints;
    //struct addrinfo *res = NULL;
    //struct addrinfo *ai = NULL;
    //char port_str[NI_MAXSERV];
    int nb_server_s = 0;
    int server_s[MAX_SERV];
    int code;
    int syslog_flag;
    time_t last_take_chld=0;

    Act_iter aend;

    Ant_mon_sensor *aparam;

    fix_time();

    log_to_syslog = 0;
    syslog_flag   = 0;
    log_name = new char[12];
    strcpy(log_name, "ant-mon");
    log_level=-1;

    // arguments parsing
    for (i = 1; i < argc; ++i) {

        // help
        if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
            printf("Usage: ant_mon [keys]\n");
            printf("       -h         help\n");
            printf("       -d         daemonize\n");
            printf("       -c file    use config file (default "
                   "/etc/antmon.conf, then ./antmon.conf)\n");
            printf("       -g N       log level N (see README) \n");
            printf("       -s         log to syslog (exludes -l,-f)\n");
            printf("       -n name    use syslog prefix 'name'\n");
            printf("       -f         log to stderr (exludes -l,-s)\n");
            printf("       -l file    log to file (excludes -f,-s)\n");
            gentle_exit(0);
        }

        // daemonize?
        if (!strcmp(argv[i], "-d")) {
            // log to syslog by default if daemon
            if(syslog_flag==0){
                log_to_syslog=1;
            }
            if(log_level<0){
                log_level=ALOG_WARN;
            }
            daemon = 1;
            continue;
        }

        // log to syslog?
        if (!strcmp(argv[i], "-s")) {
            log_to_syslog = 1;
            continue;
        }

        // log to stderr?
        if (!strcmp(argv[i], "-f")) {
            log_to_syslog=0;
            syslog_flag=1; // mark log redirection
            continue;
        }

        // new config file
        if (!strncmp(argv[i], "-c", 2)) {
            if (strlen(argv[i] + 2) > 0) {
                strncpy(conf_file, argv[i] + 2, 1023);
                conf_file[1023] = '\0';
            }
            else if (++i < argc) {
                strncpy(conf_file, argv[i], 1023);
                conf_file[1023] = '\0';
            }
            continue;
        }

        // log level
        if (!strncmp(argv[i], "-g", 2)) {
            if (strlen(argv[i] + 2) > 0) {
                sscanf(argv[i] + 2, "%d", &log_level);
            }
            else if (++i < argc) {
                sscanf(argv[i], "%d", &log_level);
            }
            continue;
        }

        // syslog name
        if (!strncmp(argv[i], "-n", 2)) {
            if (strlen(argv[i] + 2) > 0) {
                if (log_name)
                    delete log_name;
                log_name = new char[strlen(argv[i] + 2) + 1];
                strcpy(log_name, argv[i] + 2);
            }
            else if (++i < argc) {
                if (log_name)
                    delete log_name;
                log_name = new char[strlen(argv[i]) + 1];
                strcpy(log_name, argv[i]);
            }
            continue;
        }

        // log to file
        if (!strncmp(argv[i], "-l", 2)) {
            log_to_syslog = 0;
            syslog_flag = 1; //mark log redirection
            if (strlen(argv[i] + 2) > 0) {
                j = open(argv[i] + 2,
                         O_WRONLY | O_CREAT | O_APPEND | O_LARGEFILE,
                         S_IRWXU | S_IRGRP);
                if (j > 0) {
                    close(2);
                    dup2(j, 2);
                    close(j);
                }
                else {
                    LOG(ALOG_WARN | ALOG_CONF,
                        ("Cannot open %s: %s", argv[i], strerror(errno)));
                    gentle_exit(1);
                }
            }
            else if (++i < argc) {
                j = open(argv[i], O_WRONLY | O_CREAT | O_APPEND | O_LARGEFILE,
                         S_IRWXU | S_IRGRP);
                if (j > 0) {
                    close(2);
                    dup2(j, 2);
                    close(j);
                }
                else {
                    LOG(ALOG_WARN | ALOG_CONF,
                        ("Cannot open %s: %s", argv[i], strerror(errno)));
                    gentle_exit(1);
                }
            }
            continue;
        }
    }

    // set maximal log level if not a daemon
    if(log_level<0){
        log_level=ALOG_ALL;
    }
    // don't close 0,1,2 descriptors
    // but associate with /dev/null in order to prevent
    // situation when the kernel is patched
    int rd = open("/dev/null", O_RDONLY);
    int wr = open("/dev/null", O_WRONLY);
    dup2(rd, 0);
    dup2(wr, 1);

    // close stderr if log to syslog only
    if (log_to_syslog==1) {
        dup2(wr, 2);
    }


#ifdef USE_RLIMIT
    getrlimit(RLIMIT_CPU, &cpulim);
    if (cpulim.rlim_max != RLIM_INFINITY) {
        printf
            ("Warning! You have limited me by ulimit CPU. So I'll die in some time. Please, remove limits...\n");
    }
#endif

    LOG(ALOG_ALL, ("Starting..."));

    if (daemon)
        daemonize();

    signal(SIGTERM, sig_exit_handler);
    signal(SIGQUIT, sig_exit_handler);
    signal(SIGABRT, sig_exit_handler);
    signal(SIGSEGV, sig_exit_handler);
    signal(SIGXCPU, sig_exit_handler);
    signal(SIGFPE, sig_exit_handler);
    signal(SIGBUS, sig_exit_handler);

    signal(SIGPIPE, sig_log_handler);
    signal(SIGURG, sig_urg_handler);
    signal(SIGCHLD, exe_sig_chld_handler);
    signal(SIGUSR1, sig_usr1_handler);
    signal(SIGUSR2, sig_usr2_handler);
    signal(SIGHUP, sig_hup_handler);

    if (load_config("/etc/antmon.conf"))
        if (load_config("./antmon.conf")) {
            LOG(ALOG_WARN, ("Cannot open file antmon.conf"));
            gentle_exit(1);
        }

    Ant_connection::send_retr = 3;

    create_pid();

    max_interval_change = 4;    //!!! put instruction to config!!!

    nb_server_s=build_server_socket(head_port, server_s, MAX_SERV);
    if (nb_server_s < 1) {
        LOG(ALOG_WARN | ALOG_NET, ("no suitable listen address found\n"));
        gentle_exit(2);
    }

    /* this is code from Olivier Aumage, Samuel Thibault
    snprintf(port_str, NI_MAXSERV, "%u", head_port);

    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = PF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_addrlen = 0;
    hints.ai_addr = NULL;
    hints.ai_canonname = NULL;
    hints.ai_next = NULL;

    if ((code = getaddrinfo(NULL, port_str, &hints, &res))) {
        LOG(ALOG_WARN | ALOG_NET, ("Error in getaddrinfo: %s\n",
                                   gai_strerror(code)));
        if (code == EAI_SYSTEM) {
            LOG(ALOG_WARN | ALOG_NET, ("Details: %s", strerror(errno)));
        }
        gentle_exit(1);
    }

    for (ai = res; ai && nb_server_s < MAX_SERV; ai = ai->ai_next) {
        int _server_s = -1;

        if (ai->ai_socktype != hints.ai_socktype
            || ai->ai_protocol != hints.ai_protocol)
            continue;

        if ((_server_s = build_server_socket(ai)) < 0) {
            continue;
        }

        server_s[nb_server_s++] = _server_s;

    }

    if (nb_server_s < 1) {
        LOG(ALOG_WARN | ALOG_NET, ("no suitable listen address found\n"));
        gentle_exit(1);
    }

    freeaddrinfo(res);
    */
    /* end of code from Olivier Aumage, Samuel Thibault */

    LOG(ALOG_INFO | ALOG_NET,
        ("Listen on %d for others monitors", head_port));

    //
    //
    // ===================== MAIN LOOP =======================================
    //
    //
    num_of_conn = Conn_by_ip.size();

    for (;;) {
        //get current time
        fix_time();

        // call waitpid for childs sometimes...
        if((global_time-last_take_chld>CHLD_INTERVAL) ||
            kill_flag!=0){
            last_take_chld=global_time;
            take_childs();
        }

        //
        // for each agent...
        ////////////////////////////////////////////////////
        was_change = 0;
        for (i = 0; i < num_of_conn; ++i) {

            if (!Conn_by_ip[i]->enabled)
                continue;

            // randomize interval increment
            // for ALL parameters on single ip!
            interval_add = random() * max_interval_change / RAND_MAX;

            // connect if not yet...
            if (Conn_by_ip[i]->connected == 0) {

                // wait for right reconnection time?
                if(Conn_by_ip[i]->next_connect>global_time)
                    continue;

                LOG(ALOG_DEBUG2,(
                    ">>try to reconnect to agent %s",
                    Conn_by_ip[i]->name.c_str()));

                // check for timed out params
                if (!Conn_by_ip[i]->timed_out) {
                    if (Conn_by_ip[i]->disconnected_time
                        + recon_timeout > global_time) {

                        process_conn_error(i);
                        Conn_by_ip[i]->timed_out = 1;
                        Conn_by_ip[i]->disconnect();

                        // mark all associated params as timed out
                        max = Sensors_by_ip[i]->size();
                        for (j = 0; j < max; ++j) {
                            aparam = (*(Sensors_by_ip[i]))[j];

                            // is it foreign and not controlled by me?
                            if (!aparam->is_own && aparam->controlled == 'n')
                                continue;

                        }       // all associated params now marked as timed out
                    }
                }

                // try to connect
                j=Conn_by_ip[i]->connect();
                if (j!=0) {
                    // connection failed or in progress
                    if(j<0){
                        // failed... Wait some time
                        Conn_by_ip[i]->next_connect =
                            global_time + recon_interval;
                        continue;
                    }
                }

                {
                    Ant_connection *conn = Conn_by_ip[i];
                    LOG(ALOG_NET,
                        ("Socket %d: connected to %s, %d", conn->fd,
                        conn->name.c_str(), Conn_by_ip[i]->connected));
                }
                process_connected(i);

                // connected = true
            }
            else {

                // read messages
                while (Conn_by_ip[i]->getMessage()) {
                    process_message(Conn_by_ip[i]);
                    Conn_by_ip[i]->clean();
                }

                // test associated parameters
                max = Sensors_by_ip[i]->size();
                for (j = 0; j < max; ++j) {
                    aparam = (*(Sensors_by_ip[i]))[j];

                    // is it foreign and not controlled by me?
                    if (!aparam->is_own && aparam->controlled == 'n')
                        continue;

                    // is param id resolved?
                    if (aparam->id < 0) {

                        // send resolve request
                        id = new_id();
                        snprintf(tmpbuf, MAX_BUF, "%ld:a:%s\n\n", id,
                                 aparam->name.c_str());
                        tmpbuf[MAX_BUF - 1] = '\0';
                        Conn_by_ip[i]->sendMessage(tmpbuf);
                        aparam->id = 0;

                        // remember it
                        params_by_id[id] = (void *) aparam;

                        // mark it...
                        aparam->data_sent = global_time;
                        aparam->next_check =
                            global_time + aparam->interval + interval_add;
                        LOG(ALOG_DEBUG | ALOG_SENS,
                            ("next check in %d (add): %s",
                             aparam->interval + interval_add,
                             aparam->name.c_str()));
                    }
                    // is resolve request sent?
                    else if (aparam->id == 0) {
                        continue;
                    }

                    // check state
                    switch (aparam->state) {

                    case Fail_dep:

                        // check, was dependence state changed?
                        if (!was_change)
                            break;

                        // check dependencies...
                        flag = 1;
/*!!!!!!!!depends
            max=aparam->depends.size();
            for(j=0; j<max; ++j){
              if(((aparam->depends)[j])->state != Active){
                flag=0;
                break;
              }
            }
*/
                        // no, there are failed dependencies
                        if (flag)
                            break;

                        // all dependencies are good!
                        // restore old state
                        aparam->state = aparam->old_state;
                        // reset flag
                        aparam->data_sent = 0;
                        //and continue checking!

                    case Active:
                    case Fail_min:
                    case Fail_max:
                    case Fail_get:

                        // is request sent already?
                        if (aparam->data_sent)
                            continue;

                        // do not check yet...
                        if (aparam->next_check > global_time)
                            break;

                        // send request
                        id = new_id();
                        snprintf(tmpbuf, MAX_BUF, "%ld:g:%ld\n\n", id,
                                 aparam->id);
                        tmpbuf[MAX_BUF - 1] = '\0';
                        Conn_by_ip[i]->sendMessage(tmpbuf);

                        // remember it
                        params_by_id[id] = (void *) aparam;

                        // mark it...
                        aparam->data_sent = global_time;
                        aparam->next_check =
                            global_time + aparam->interval + interval_add;
                        LOG(ALOG_DEBUG | ALOG_SENS,
                            ("next check in %d seconds: %s",
                             aparam->interval + interval_add,
                             aparam->name.c_str()));

                        //push to awaited...
                        Waited_params.push_back(aparam);

                        break;

                    case Fail_tmout:

                        // check again?
                        if (aparam->next_check < global_time) {

                            // send request...
                            id = new_id();
                            snprintf(tmpbuf, MAX_BUF, "%ld:p:\n\n", id);
                            tmpbuf[MAX_BUF - 1] = '\0';
                            Conn_by_ip[i]->sendMessage(tmpbuf);

                            // remember it
                            params_by_id[id] = (void *) aparam;

                            // mark it...
                            aparam->data_sent = global_time;
                            aparam->next_check =
                                global_time + aparam->interval + interval_add;
                            LOG(ALOG_DEBUG | ALOG_SENS,
                                ("next check in %d: %s",
                                 aparam->interval + interval_add,
                                 aparam->name.c_str()));

                            //push to awaited...
                            Waited_params.push_back(aparam);
                        }

                        break;

                    case Disabled:

                        break;

                    default:
                        LOG(ALOG_WARN,
                            ("Bad state... %s (%s)", aparam->conn->name.c_str(),
                             aparam->name.c_str()));
                    }
                }

                // actually send all
                Conn_by_ip[i]->flush();

                // is there error?
                if (Conn_by_ip[i]->error) {

                    // save name
                    const char *name = Conn_by_ip[i]->name.c_str();
                    LOG(ALOG_WARN | ALOG_SENS,
                        ("Error on %s (%s)", name, strerror(errno)));

                    // save error
                    int err = Conn_by_ip[i]->error;

                    //Conn_by_ip[i]->name = NULL; // do not free it...
                    Conn_by_ip[i]->finalize();

                    // reinitialize
                    Conn_by_ip[i]->init(name);  // restore name
                    Conn_by_ip[i]->next_connect =
                        global_time + recon_interval;
                    Conn_by_ip[i]->error = err;
                    process_conn_error(i);
                }
            }
        }                       //for all connections

        // check for timed out parameters
        std::list < class Ant_mon_sensor * >::iterator end =
            Waited_params.end();
        for (std::list < class Ant_mon_sensor * >::iterator li =
             Waited_params.begin(); li != end;) {

            time_t ds = (*li)->data_sent;
            time_t tm = (*li)->timeout;
            // timed out!!!
            if (ds + tm < global_time) {

                // react
                process_timed_out(*li);

                // erase list entry
                std::list < class Ant_mon_sensor * >::iterator del = li++;
                Waited_params.erase(del);
            }
            else
                ++li;
        }

        // flush all actions
        aend = Actions.end();
        for (Act_iter ali = Actions.begin(); ali != aend; ++ali) {
            ali->second->flush();
        }

        //
        // try to accept new connections from other monitors
        //
        {
            struct sockaddr addr;
            socklen_t len;

            mon_sd = multi_accept(server_s, nb_server_s, &addr, &len);

            if (mon_sd >=0) {
                char sport[NI_MAXSERV];
                // accepted...
                /*serv_name = inet_ntoa(addr.sin_addr); */
                //get_name_and_port(mon_sd, serv_name, sport);
                if ((code = getnameinfo(&addr, len, serv_name,
                                        NI_MAXHOST, sport, NI_MAXSERV,
                                        NI_NUMERICHOST|NI_NUMERICSERV))) {
                    LOG(ALOG_WARN | ALOG_NET, ("Error on getaddrinfo: %s",
                                               gai_strerror(code)));
                    if (code == EAI_SYSTEM) {
                        LOG(ALOG_WARN | ALOG_NET,
                        ("Details: %s", strerror(errno)));
                    }
                    strcpy(serv_name,"undefined_host");
                }

                LOG(ALOG_DEBUG2,("Accepted %s:%s",serv_name,sport));

                std::list < std::string * >::iterator hli, hlend;
                // find it in the table
                for (i = 0; i < heads.size(); ++i) {
                    LOG(ALOG_DEBUG2,("Checking!\n"));
                    if (heads[i]->name == serv_name){
                        flag=1;
                        break;
                    }
                    hlend = heads[i]->ips.end();
                    flag = 0;
                    for (hli = heads[i]->ips.begin(); hli != hlend; ++hli) {
                        LOG(ALOG_DEBUG2,("Check %s = %s",
                            (**hli).c_str(),serv_name));
                        if (**hli == serv_name) {
                            flag = 1;
                            break;
                        }
                    }
                    if (flag)
                        break;
                }

                // found it?
                if (i < heads.size()) {

                    // set apropriate parameters for conenction...
                    heads[i]->in_conn->init(serv_name);
                    heads[i]->in_conn->fd = mon_sd;
                    if (fcntl
                        (mon_sd, F_SETFL,
                         O_NONBLOCK | fcntl(mon_sd, F_GETFL, O_NONBLOCK)) != 0) {
                        LOG(ALOG_WARN | ALOG_NET,
                            ("Cannot make port %d nonblocked (%s)", mon_sd,
                             strerror(errno)));
                        gentle_exit(5);
                    }
                    heads[i]->in_conn->connected = 1;
                    heads[i]->last_ping = global_time;
                    LOG(ALOG_NET,
                        ("Got connection from head %s (%s:%s). %d-th head. fd=%d",
                         heads[i]->name.c_str(), serv_name, sport, i, mon_sd));
                }
                else {
                    close(mon_sd);
                    LOG(ALOG_WARN | ALOG_NET,
                        ("Illegal connection from %s. Ignore.", serv_name));
                }
            }
        }

        //
        // communicate with head servers
        ///////////////////////////////////////////////////////////////////////
        for (i = 0; i < ANT_MAX_HEADS; ++i) {

/*
      // is it myself?
      if(i==my_head_index)
        continue;
*/

            // last?
            if (head_outs[i] < 0)
                break;

            // Needs init?
            if (heads[head_outs[i]]->state == Head_No_Init) {
                int index = head_outs[i];

                /* change state */
                heads[index]->state = Head_Fail;
                heads[index]->last_ping = global_time;

                heads[index]->out_conn = NULL;
                //heads[index]->out_conn_list = new Ant_connection *[2];

                heads[index]->out_conn_list.push_back(new Ant_connection);
                heads[index]->out_conn_list[0]->
                    init(heads[head_outs[i]]->name.c_str());
                heads[index]->out_conn_list[0]->port = head_port;

                // fill head names...
                {
                    struct addrinfo hints;
                    struct addrinfo *result, *rp;
                    int s;
                    char name[NI_MAXHOST];

                    memset(&hints, 0, sizeof(struct addrinfo));
                    hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
                    hints.ai_socktype = SOCK_STREAM;
                    hints.ai_flags = AI_ALL|AI_V4MAPPED|
                                     AI_CANONNAME|AI_PASSIVE;
                                     /* For wildcard IP address */
                    hints.ai_protocol = 0;          /* Any protocol */

                    s = getaddrinfo(heads[index]->name.c_str(), NULL,
                                    &hints, &result);
                    if(s==0){
                        for(rp=result; rp!=NULL; rp=rp->ai_next){
                            if(getnameinfo(rp->ai_addr, rp->ai_addrlen,
                                    name, NI_MAXHOST, NULL, 0,
                                    NI_NUMERICHOST) == 0){
                                LOG(ALOG_DEBUG2, ("head[%d] name: %s\n",
                                    index,name));
                                heads[index]->ips.push_front(
                                    new std::string(name)
                                );
                               }
                        }
                        freeaddrinfo(result);
                    }
                    else{
                        LOG(ALOG_WARN | ALOG_NET,(
                            "getnameinfo: %s\n", gai_strerror(s)));
                    }
                }
                heads[index]->out_conn_list[1] = NULL;
                heads[index]->next_ping = global_time;

            }                   // ~ needs init

            //
            else if (heads[head_outs[i]]->state == Head_Fail) {

                if(heads[head_outs[i]]->timed_out){
                    // check if it is not time to reconnect yet
                    if(heads[head_outs[i]]->next_ping>global_time)
                        continue;
                }
                LOG(ALOG_DEBUG2, (">>head %s fails",
                            heads[head_outs[i]]->name.c_str()));
                if (heads[head_outs[i]]->out_conn == NULL) {
                    // check heads[head_outs[i]].out_conn_list
                    // select real one...

                    for (j = 0; heads[head_outs[i]]->out_conn_list[j] != NULL;
                         ++j) {
                        // try to connect
                        LOG(ALOG_DEBUG2,
                            (">>try to connect to head %s",
                            heads[head_outs[i]]->name.c_str()));
                        int status = heads[head_outs[i]]->out_conn_list[j]->
                            connect();

                        if (status < 0) {

                            // error!
                            if (!heads[head_outs[i]]->timed_out) {
                                if (heads[head_outs[i]]->last_ping +
                                    head_tmout < global_time) {

                                    heads[head_outs[i]]->timed_out = 1;

                                    /* connecting timed out */
                                    heads[head_outs[i]]->out_conn =
                                        heads[head_outs[i]]->out_conn_list[j];
                                    LOG(ALOG_WARN | ALOG_NET,
                                        ("Head %s timed out.",
                                         heads[head_outs[i]]->name.c_str()));

                                    // do the action!
                                    process_conn_error_head(head_outs[i], 0);
                                }
                            }
                            heads[head_outs[i]]->next_ping =
                                global_time + head_ping;
                            continue;
                        }
                        else if (status > 0) {
                            // connecion in progress
                            LOG(ALOG_DEBUG2, (">>Connection is in progress now..."));
                            continue;
                        }

                        // connected!!!
                        heads[head_outs[i]]->out_conn =
                            heads[head_outs[i]]->out_conn_list[j];
                        LOG(ALOG_NET,
                            ("Connected to head %s",
                             heads[head_outs[i]]->name.c_str()));
                        heads[head_outs[i]]->state = Head_Active;
                        heads[head_outs[i]]->next_ping =
                            global_time + head_ping;
                        process_connected_head(head_outs[i]);
                        break;
                    }
                }
                else {
                    // already was connected. Just reconnect
                    LOG(ALOG_DEBUG2, (">>try to reconnect head (%s)",
                                heads[head_outs[i]]->name.c_str()));
                    int status =
                        heads[head_outs[i]]->out_conn->connect();

                    if (status == 0) {
                        // connected!!!

                        LOG(ALOG_NET,
                            ("Reconnected to head %s",
                             heads[head_outs[i]]->name.c_str()));
                        heads[head_outs[i]]->state = Head_Active;
                        heads[head_outs[i]]->next_ping =
                            global_time + head_ping;
                        heads[head_outs[i]]->timed_out = 0;
                        process_connected_head(head_outs[i]);
                    }
                    else {
                        if(status > 0){
                            // connection in progress
                            LOG(ALOG_DEBUG2, (">>Connection to head %s in progress now...",
                            heads[head_outs[i]]->name.c_str()));
                            continue;
                        }
                        // connection failed
                        heads[head_outs[i]]->next_ping =
                            global_time + head_ping;
                        continue;
                    }
                }
            }
            else if (heads[head_outs[i]]->state == Head_Active) {

                LOG(ALOG_NET,
                    (">> %s - Active", heads[head_outs[i]]->name.c_str()));
                // next ping?
                if (heads[head_outs[i]]->next_ping < global_time) {
                    LOG(ALOG_NET,
                        ("PING to %s (%d-th head)",
                         heads[head_outs[i]]->name.c_str(), head_outs[i]));
                    snprintf(tmpbuf, MAX_BUF, "%d:h\n\n", my_head_index);
                    tmpbuf[MAX_BUF - 1] = '\0';
                    heads[head_outs[i]]->out_conn->sendMessage(tmpbuf);
                    heads[head_outs[i]]->next_ping = global_time + head_ping;

                    // actually send all
                    heads[head_outs[i]]->out_conn->flush();
                }

                // is there error?
                if (heads[head_outs[i]]->out_conn->error) {

                    // save name
                    const char *name = heads[head_outs[i]]->out_conn->name.c_str();
                    LOG(ALOG_WARN | ALOG_NET, ("Error on server %s", name));

                    // save error
                    //int err = heads[head_outs[i]]->out_conn->error;

                    heads[head_outs[i]]->out_conn->finalize();

                    // reinitialize
                    process_conn_error_head(head_outs[i], 0);
                }
            }
            else {
                LOG(ALOG_WARN | ALOG_NET,
                    ("Bad state of head server connection - %d (%s)\n", i,
                     heads[head_outs[i]]->name.c_str()));
            }
        }                       //for all head servers

        //
        //
        // check timed out head servers
        // and receive data from
        //
        for (i = 0; i < ANT_MAX_HEADS; ++i) {
            // last?
            if (head_ins[i] == -1)
                break;

            //      LOG(ALOG_WARN,("Check input from %d\n",head_ins[i]));

            if (heads[head_ins[i]]->state != Head_Active)
                break;

            //      LOG(ALOG_WARN,("Check input from %d\n",head_ins[i]));

            // timed out?
            if (!heads[head_outs[i]]->timed_out) {
                if (heads[head_ins[i]]->last_ping + head_tmout < global_time) {
                    const char *name = heads[head_ins[i]]->in_conn->name.c_str();
                    LOG(ALOG_WARN | ALOG_NET, ("Head %s timed out...", name));

                    heads[head_outs[i]]->timed_out = 1;
                    if (heads[head_ins[i]]->in_conn)
                        heads[head_ins[i]]->in_conn->finalize();
                    if (heads[head_ins[i]]->out_conn)
                        heads[head_ins[i]]->out_conn->finalize();
                    heads[head_ins[i]]->state = Head_Fail;

                    // do the action!
                    process_conn_error_head(head_ins[i], 0);
                }
            }

            // read messages
            while (heads[head_ins[i]]->in_conn->getMessage()) {
                //LOG(ALOG_WARN,("Ping from %s?\n",heads[head_ins[i]]->name.c_str()));
                process_message_head(heads[head_ins[i]]->in_conn);
                heads[head_ins[i]]->in_conn->clean();
                heads[head_ins[i]]->last_ping = global_time;
            }
        }

        // sleep
        sleep_time.tv_sec = 0;
        sleep_time.tv_usec = 1000000;
        select(0, NULL, NULL, NULL, &sleep_time);

    }                           // main loop
    return 1;
}




void sig_log_handler(int sig)
{
    LOG(ALOG_WARN, ("Signal %d received. Ignore.", sig));
}

void sig_urg_handler(int sig)
{
    LOG(ALOG_WARN, ("Signal %d received. Ignore.", sig));
}

void sig_exit_handler(int sig)
{
    LOG(ALOG_WARN, ("Signal %d received. Exit now.", sig));

    gentle_exit(0);
}

/*
 *   Create dump file to /tmp/ant_mon.dump
 *
 */
void sig_usr1_handler(int sig)
{
    FILE *f;
    int i, j, size, size2;
    class Ant_head *hptr;
    class Ant_connection *aptr;
    class Ant_mon_sensor *pptr;
    class vector < Ant_mon_sensor * >*vptr;
    char buf[64];
    struct tm *tm;

    LOG(ALOG_WARN, ("Signal USR1 received."));


    if ((f = fopen("/tmp/ant_mon.dump", "w")) == NULL) {
        LOG(ALOG_WARN, ("Cannot create /tmp/ant_mon.dump."));
        return;
    }

    /* dump status */
    fprintf(f, "Head port: %d\nAgent port: %d\n", head_port, agent_port);

    fprintf(f, "Heads:\n");
    size = heads.size();
    for (i = 0; i < size; ++i) {
        hptr = heads[i];
        fprintf(f, "%c%2d: %16s : %s; ",
                (i==my_head_index)?'*':' ',
                i+1,
                hptr->name.c_str(),
                (hptr->state == Head_Active ? "active" :
                 (hptr->state == Head_Fail ? "failed" :
                  (hptr->state == Head_Disabled ? "disabl" :
                   (hptr->state == Head_No_Init ? "noinit" : "ERROR!")
                  )
                 )
                )
            );
        /* last ping: */
        tm = localtime(&hptr->last_ping);
        strftime(buf, 63, "%F %H:%M:%S", tm);
        fprintf(f, "last ping: %s; ", buf);
        /* next ping: */
        tm = localtime(&hptr->next_ping);
        strftime(buf, 63, "%F %H:%M:%S", tm);
        fprintf(f, "next ping: %s\n", buf);
    }

    fprintf(f, "My index: %d\n", my_head_index+1);

    fprintf(f, "Dead heads indexes:\n");
    for (i = 0; i < ANT_MAX_HEADS; ++i) {
        if (dead_heads[i] < 0)
            break;
        fprintf(f, " %hd;", dead_heads[i]+1);
    }
    fprintf(f, "\n");

    fprintf(f, "Pings me indexes:\n");
    for (i = 0; i < ANT_MAX_HEADS; ++i) {
        if (head_ins[i] < 0)
            break;
        fprintf(f, " %hd;", head_ins[i]+1);
    }
    fprintf(f, "\n");

    fprintf(f, "Pinging by me indexes:\n");
    for (i = 0; i < ANT_MAX_HEADS; ++i) {
        if (head_outs[i] < 0)
            break;
        fprintf(f, " %hd;", head_outs[i]+1);
    }
    fprintf(f, "\n");

    fprintf(f, "Connections:\n");
    fprintf(f, "    id:         name           :conn:enab:state: next_conn\n");
    size = Conn_by_ip.size();
    for (i = 0; i < size; ++i) {
        aptr = Conn_by_ip[i];
        fprintf(f, "%6d:%24s:%s:%s:%s:", i, aptr->name.c_str(),
                aptr->connected ? " ok " : " no ",
                aptr->enabled ? " ena" : "disa",
                aptr->error ? "error" : "  ok ");

        /* next connect: */
        tm = localtime(&aptr->next_connect);
        strftime(buf, 63, "%F %H:%M:%S", tm);
        fprintf(f, " %s\n", buf);
    }

    fprintf(f, "Params:            "
           "(b=back required/c=back current, r=ret required/c=ret current) \n");
    fprintf(f,"    id:              name              :"
              "    conn. name  :state:own:ctr: value  :"
              "    min  :   max  :b/c:r/c: next check\n");

    size = Sensors_by_ip.size();
    for (i = 0; i < size; ++i) {
        vptr = Sensors_by_ip[i];
        size2 = vptr->size();
        for (j = 0; j < size2; ++j) {
            pptr = (*vptr)[j];

            fprintf(f,
            "%6ld:%32s:%16s:%s:%s:%s:%8.2lg:%8.2lg:%8.2lg:%1d/%1d:%1d/%1d:",
                    pptr->id, pptr->name.c_str(), pptr->conn->name.c_str(),
                    (pptr->state == Active ?
                    "  OK " : (pptr->state == Fail_min ?
                    "X min" : (pptr->state == Fail_max ?
                    "X max" : (pptr->state == Fail_dep ?
                    "X dep" : (pptr->state == Fail_tmout ?
                    "tmout" : (pptr->state == Fail_get ?
                    "X get" : (pptr->state == Disabled ?
                    "Disab" : "ERROR"))))))),
                    pptr->is_own ? " my" : " no",
                    pptr->controlled ? "yes" : "not", pptr->value,
                    pptr->min_value, pptr->max_value, pptr->bad_count,
                    pptr->bad_count_cur, pptr->ret_count,
                    pptr->ret_count_cur);

            /* next connect: */
            tm = localtime(&pptr->next_check);
            strftime(buf, 63, "%F %H:%M:%S", tm);
            fprintf(f, " %s\n", buf);
        }
    }
    fclose(f);
}

void sig_usr2_handler(int sig)
{
    LOG(ALOG_WARN, ("Signal USR2 received."));
}

/*
 * Reload config file
 *
 */
void sig_hup_handler(int sig)
{
    LOG(ALOG_WARN, ("Signal HUP received."));
    config_reload();

}


int GetNonblockSocket()
{
    int sock;
    if ((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
        LOG(ALOG_WARN | ALOG_NET,
            ("Error creating socket: %s", strerror(errno)));
        return (-1);
    }
    if (fcntl(sock, F_SETFL, O_NONBLOCK) != 0) {
        LOG(ALOG_WARN | ALOG_NET,
            ("Error making socket nonblocked: %s", strerror(errno)));
        return (-1);
    }
    return sock;
}

// cancel id waiting...
void id_not_awaited_now(long int id)
{

    Ant_mon_sensor *param = (Ant_mon_sensor *) params_by_id[id];

    // delete parameter from hash
    params_by_id.erase(id);

    // not found?
    if (param == NULL)
        return;

    // and from awaited
    for (std::list < class Ant_mon_sensor * >::iterator li =
         Waited_params.begin(); li != Waited_params.end();) {

        // match?
        if ((*li) == param) {
            Waited_params.erase(li);
            break;
        }
        else
            ++li;
    }
}

// message from head server
void process_message_head(Ant_connection * conn)
{
    char *p = conn->buffer;
    int from, subject;

    // where it from?
    from = 0;
    while (*p && (*p >= '0' && *p <= '9')) {
        from *= 10;
        from += *p - '0';
        ++p;
    }

    // skip ':'
    if (*p == '\0' || *(++p) == '\0') {
        LOG(ALOG_WARN | ALOG_NET, ("Bad message from another head server\n"));
        conn->clean();
        return;
    }

    // read message type
    switch (*p) {
    case 'r':
        // resurrected server
        if (*(++p) == '\0') {
            LOG(ALOG_WARN | ALOG_NET,
                ("Bad message from head server '%s' about resurrection\n",
                 conn->name.c_str()));
            conn->clean();
            return;
        }
        // read resurrected server number
        subject = 0;
        while (*p && (*p >= '0' && *p <= '9')) {
            subject *= 10;
            subject += *p - '0';
            ++p;
        }

        if (*p == '\0') {
            LOG(ALOG_WARN | ALOG_NET,
                ("Bad message from head server '%s' about resurrection of %d\n",
                 conn->name.c_str(), subject));
            conn->clean();
            return;
        }

        // react!
        // test if it is me (hmmm) or already resurrected server
        if (subject == my_head_index)
            return;

        process_connected_head(subject);
        break;
    case 'f':
        // failed server
        if (*(++p) == '\0') {
            LOG(ALOG_WARN | ALOG_NET,
                ("Bad message from head server '%s' about failure\n",
                 conn->name.c_str()));
            conn->clean();
            return;
        }
        // read failed server number
        subject = 0;
        while (*p && (*p >= '0' && *p <= '9')) {
            subject *= 10;
            subject += *p - '0';
            ++p;
        }

        if (*p == '\0') {
            LOG(ALOG_WARN | ALOG_NET,
                ("Bad message from head server '%s' about failure on of %d\n",
                 conn->name.c_str(), subject));
            conn->clean();
            return;
        }

        // react!
        // test if it is me (hmmm) or already resurrected server
        if (subject == my_head_index)
            return;

        process_conn_error_head(subject, 1);
        break;
    case 'h':
        // heartbeat
        heads[from]->last_ping = global_time;
        break;

    default:
        // error message
        break;
    }
}                               // process_message_head

// returns index of head server, which must control this parameter
int active_head(Ant_mon_sensor * p)
{
    int i, j, flag;

    for (i = 0; i < ANT_MAX_HEADS; ++i) {
        //    LOG(ALOG_DEBUG2,(">> test head %d",p->heads[i]));
        if (p->heads[i] < 0)
            return -1;          // No more heads can control it!

        flag = 1;
        for (j = 0; j < ANT_MAX_HEADS; ++j) {
            //      LOG(ALOG_DEBUG2,(">> dead head[%d]=%d",j,dead_heads[j]));
            if (dead_heads[j] < 0)
                break;
            if (dead_heads[j] == p->heads[i]) {
                flag = 0;       // found in dead heads list
                break;
            }
        }
        if (flag)               // Not found in dead heads list
            return p->heads[i];
    }
    return -1;
}

// connection with head server was established
void process_connected_head(int i)
{
    int j, param, max, max_param, head, not_found;
    Ant_mon_sensor *p;

    // check if it is already processed and delete it from deads
    not_found = 1;
    for (j = 0; j < ANT_MAX_HEADS; ++j) {

        //    LOG(ALOG_DEBUG2,("DEADLIST %d: %d ?= %d",j,dead_heads[j],i));
        // found end of list?
        if (dead_heads[j] < 0)
            break;

        // found it?
        if (dead_heads[j] == i) {

            not_found = 0;

            //shift the rest
            for (; j < ANT_MAX_HEADS - 1; ++j) {
                dead_heads[j] = dead_heads[j + 1];

                // last?
                if (dead_heads[j] == -1)
                    break;
            }
            // mark the end directly (in case of ful list)
            dead_heads[ANT_MAX_HEADS - 1] = -1;

            break;
        }
    }
    if (not_found)
        return;

    // check, if some sensors is not on our competention now...
    max_param = Sensors_by_ip.size();
    for (param = 0; param < max_param; ++param) {
        max = Sensors_by_ip[param]->size();
        for (j = 0; j < max; ++j) {
            p = (*(Sensors_by_ip[param]))[j];

            // check if sensor was on our care
            if (p->controlled == 'y') {
                head = active_head(p);
                //        LOG(ALOG_DEBUG2,(">> process_connected: sensor controlled by %d\n",head));

                if (head != my_head_index) {
                    // Release it...
                    LOG(ALOG_DEBUG | ALOG_SENS,
                        ("Sensor '%s' on '%s' was on my care. Release it to %d head (conn_head)\n",
                         p->name.c_str(), p->conn->name.c_str(), head));
                    p->controlled = 'n';
                }
                // no, it is owned by another dead server...
            }
        }
    }

    // send other monitors message about resurrection
    for (i = 0; i < ANT_MAX_HEADS; ++i) {
        if (head_outs[i] == -1)
            break;

        // create a message (resurrect)
        snprintf(tmpbuf, MAX_BUF, "%d:r:%d\n\n", my_head_index, i);
        tmpbuf[MAX_BUF - 1] = '\0';
        heads[head_outs[i]]->out_conn->sendMessage(tmpbuf);
    }
}                               // process_connected_head

// connection with head server was lost
void process_conn_error_head(int i, int by_message)
{
    int j, param, max, max_param, head;
    Ant_mon_sensor *p;

    // check if it is already processed and add it to deads
    for (j = 0; j < ANT_MAX_HEADS; ++j) {

        //    LOG(ALOG_DEBUG2,("ERR_DEADLIST[%d]: %d ?= %d",j,dead_heads[j],i));
        // already marked as dead?
        if (dead_heads[j] == i)
            break;

        // found end of list?
        if (dead_heads[j] < 0) {
            dead_heads[j] = i;

            // mark the end
            ++j;
            if (j < ANT_MAX_HEADS)
                dead_heads[j] = -1;
            break;
        }
    }

    // check, if some sensors is to be on our competention now...
    max_param = Sensors_by_ip.size();
    for (param = 0; param < max_param; ++param) {
        max = Sensors_by_ip[param]->size();
        for (j = 0; j < max; ++j) {
            p = (*(Sensors_by_ip[param]))[j];

            // skip our sensors
            if (p->is_own)
                continue;

            // skip if sensor was on our care already
            if (p->controlled == 'y')
                continue;


            head = active_head(p);
            LOG(ALOG_WARN | ALOG_NET,
                ("ERR HEAD: %d (%d)", head, my_head_index));
            if (head == my_head_index) {
                // Now it is on my care
                LOG(ALOG_WARN | ALOG_SENS,
                    ("Sensor '%s' on '%s' is on my care now instead of %d. (err head)\n",
                     p->name.c_str(), p->conn->name.c_str(), i));
                p->controlled = 'y';
            }
        }
    }

    // send other monitors message about failure
    for (j = 0; j < ANT_MAX_HEADS; ++j) {
        if (head_outs[j] == -1)
            break;

        // create a message (failure)
        snprintf(tmpbuf, MAX_BUF, "%d:f:%d\n\n", my_head_index, i);
        tmpbuf[MAX_BUF - 1] = '\0';
        if (heads[head_outs[j]]->out_conn) {
            heads[head_outs[j]]->out_conn->sendMessage(tmpbuf);
            heads[head_outs[j]]->out_conn->flush();
        }
    }

    //react!
    on_any_head_death(i);

    if (by_message)
        return;

    // react!
    on_head_death(i);
}                               // process_conn_error_head


// message from agent
void process_message(Ant_connection * conn)
{
    long int id;
    int i, fail_dep, bad_val;
    Ant_mon_sensor *param;

    // read id
    id = 0;
    for (i = 0; conn->buffer[i] >= '0' && conn->buffer[i] <= '9'; ++i) {
        id = id * 10 + conn->buffer[i] - '0';
    }

    // check format
    if (conn->buffer[i] != ':' || conn->buffer[i + 2] != ':') {
        LOG(ALOG_WARN | ALOG_NET,
            ("Bad answer from %s (%s)", conn->name.c_str(), conn->buffer));
        return;
    }
    ++i;

    LOG(ALOG_INFO, ("ANSWER from '%s': '%s'\n", conn->name.c_str(),
                    conn->buffer));
    // read action
    switch (conn->buffer[i]) {
        // answer on 'get' request
    case 'g':

        // find request
        param = (Ant_mon_sensor *) params_by_id[id];

        // not found?
        if (param == NULL) {
            params_by_id.erase(id);
            LOG(ALOG_SENS,
                ("g: id not exists from %s (%s)",
                 conn->name.c_str(), conn->buffer));
            return;
        }

        // not resolved?
        if (param->id < 1) {
            params_by_id.erase(id);
            LOG(ALOG_SENS,
                ("g: unexpected from %s (%s)",
                conn->name.c_str(), conn->buffer));
            return;
        }
        // read value
        i += 2;

        // check if bad value
        if (conn->buffer[i] == 'x') {
            bad_val = 1;
        }
        else {
            bad_val = 0;
            float value;
            long int interval=global_time-param->last_time;
            if(param->last_time==0)
                interval=0;

            param->last_time=global_time;
            if(param->type==Ant_Sensor_String){
                param->strvalue=conn->buffer + i + 1;
            }
            else{
                sscanf(conn->buffer + i, "%f", &value);
                if(interval==0){
                    param->rate=0;
                }
                else{
                    param->rate=(value-param->value)/interval;
                }
                param->value=value;
            }
            if(interval==0){
                param->rate=0;
            }
            else{
                param->rate=(value-param->value)/interval;
            }
        }

        // rewind to end
        while (conn->buffer[i] != '\n') {
            ++i;
        }

        // save Fail_dep state, if needed
        fail_dep = (param->state == Fail_dep) ? 1 : 0;

        // make check for new values/rates...
        if(test_param(param, id, bad_val)){
            LOG(ALOG_WARN | ALOG_SENS, ("Ooops: Bad state in %s (%s)",
                conn->name.c_str(), conn->buffer));
        }


        param->data_sent = 0;
        //param->next_check=global_time+aparam->interval+interval_add;

        break;                  //type=get

        // answer on 'add' request (resolve)
    case 'a':

        // find request
        param = (Ant_mon_sensor *) params_by_id[id];

        // not found?
        if (param == NULL) {
            params_by_id.erase(id);
            LOG(ALOG_SENS,
                ("a: id not exists from %s (%s)",
                conn->name.c_str(), conn->buffer));
            return;
        }

        {
            // read new parameter id
            long int param_id = 0;

            // not registered?
            if (conn->buffer[i + 2] == 'x') {
                id_not_awaited_now(id);

                LOG(ALOG_SENS,
                    ("Parameter %s not registered on %s", param->name.c_str(),
                     param->conn->name.c_str()));

                // DISABLE IT!!!!
                param->state = Disabled;

            }

            for (i += 2; conn->buffer[i] >= '0' && conn->buffer[i] <= '9';
                 ++i) {
                param_id = param_id * 10 + conn->buffer[i] - '0';
            }

            // initialize parameter
            param->id = param_id;
            param->data_sent = 0;
            if (param->state != Disabled && param->state != Fail_dep)
                param->state = Active;

            id_not_awaited_now(id);

            param->next_check = global_time + 1;

            if (param->id)
                LOG(ALOG_SENS,
                    ("Parameter %s resolved to %ld", param->name.c_str(),
                     param->id));

            param->next_check = 0;      // get parameter now!
        }
        break;                  // type=add

        // answer on 'ping' request
    case 'p':

        // find request
        param = (Ant_mon_sensor *) params_by_id[id];

        // not found?
        if (param == NULL) {
            params_by_id.erase(id);
            LOG(ALOG_SENS,("p: id not exists from %s (%s)",
                            conn->name.c_str(), conn->buffer));
            return;
        }

        // check, if really pong
        if (param->state != Fail_tmout) {
            LOG(ALOG_NET,
                ("Pong from not timed out parameter. Ignore. (%s %s)",
                 conn->name.c_str(), param->name.c_str()));
            break;
        }

        // set proper status
        if (bad_val) {
            param->state = Fail_get;
        }
        else if (param->value < param->min_value) {
            param->state = Fail_min;
        }
        else if (param->value > param->max_value) {
            param->state = Fail_max;
        }
        else {
            param->state = Active;
        }

        param->data_sent = 0;
        param->next_check = 0;  // get parameter now!
        LOG(ALOG_DEBUG, ("next check now... (%s)", param->name.c_str()));

        break;                  // type=ping

        // all other types are invalid
    default:
        LOG(ALOG_WARN | ALOG_SENS,
            ("Bad answer type from %s (%s)", conn->name.c_str(), conn->buffer));
    }

}

/***************************************
 *
 * Do parameter check on new value
 *
 * *************************************/

int test_param(Ant_mon_sensor *param, long int id, int bad_val){

    int  fail_dep = (param->state == Fail_dep) ? 1 : 0;

    // test param!
    switch (param->state) {
    case Disabled:
        id_not_awaited_now(id);
        break;

    case Fail_tmout:

        // returned back!
        param->bad_count_cur = 0;
        param->ret_count_cur = 0;

        // check value
        if (bad_val) {
            process_bad_value(param);
            on_single_bad(param);
            //on_fail(param);   !!! create on_fail_get
            param->state = Fail_get;
            param->bad_count_cur = 0;
        }
        else
        switch(param->type){
        case Ant_Sensor_Data:
        case Ant_Sensor_Counter:
            if (param->value < param->min_value) {

                // FAIL min
                param->state = Fail_min;
                process_bad_value(param);
                on_single_bad(param);
            }
            else if (param->value > param->max_value) {

                // FAIL max
                param->state = Fail_max;
                process_bad_value(param);
                on_single_bad(param);
            }
            else {

                // GOOD
                param->state = Active;
                process_good_value(param);
                on_single_good(param);
            }
            break;
        case Ant_Sensor_String:
            if(param->strvalue==param->stretalon){
                // GOOD
                param->state = Active;
                process_good_value(param);
                on_single_good(param);
            }
            else{
                // FAIL
                param->state = Fail_max;
                process_bad_value(param);
                on_single_bad(param);
            }
            break;
        }

        param->data_sent = 0;
        param->next_check = 0;      // get value now!!!
        LOG(ALOG_DEBUG, ("next check now (%s)", param->name.c_str()));

        id_not_awaited_now(id);

        break;              //get param; fail_tmout

    case Active:
    case Fail_dep:         // <-- answers are accepted anyway...

        if (bad_val) {
            process_bad_value(param);
            on_single_bad(param);
            //on_fail(param);   !!! create on_fail_get
            param->state = Fail_get;
            // reset bad counter
            param->bad_count_cur = 0;
        }
        // out of range?
        else
        switch(param->type){
        case Ant_Sensor_Data:
        case Ant_Sensor_Counter:
            if (param->value < param->min_value
                 || param->value > param->max_value) {
                process_bad_value(param);
                on_single_bad(param);

                if((param->state==Fail_max) ||
                   (param->state==Fail_min)){
                    // Good -> Fail
                    on_fail(param);
                }
            }
            else {
                // reset bad counter
                param->bad_count_cur = 0;

                on_single_good(param);
            }
            break;
        case Ant_Sensor_String:
            if(param->strvalue!=param->stretalon){
                process_bad_value(param);
                on_single_bad(param);

                if(param->state==Fail_max){
                    // Good -> Fail
                    on_fail(param);
                }
            }
            break;
        }

        id_not_awaited_now(id);

        // restore Fail_dep state, if needed
        if (fail_dep) {
            param->state = Fail_dep;
            LOG(ALOG_DEBUG, ("FAIL DEP!!!"));
        }

        break;
        //End processing Active state

    case Fail_min:
    case Fail_max:
    case Fail_get:

        if (bad_val) {
            // was it not Fail_get?
            if (param->state != Fail_get) {
                param->state = Fail_get;
                // reset good counter
                param->ret_count_cur = 0;
                param->bad_count_cur = 0;
            }
        }
        // in safe range?
        else
        switch(param->type){
            case Ant_Sensor_Data:
            case Ant_Sensor_Counter:

                if (param->value >= param->ret_min_value
                         && param->value <= param->ret_max_value) {
                    process_good_value(param);
                    on_single_good(param);

                }
                else{
                    // reset good counter
                    param->ret_count_cur = 0;
                    // act on single bad value
                    on_single_bad(param);
                }
                break;
            case Ant_Sensor_String:
                if(param->strvalue==param->stretalon){
                    process_good_value(param);
                    on_single_good(param);
                }
                else{
                    // reset good counter
                    param->ret_count_cur = 0;
                    // act on single bad value
                    on_single_bad(param);
                }
                break;
        }
        // Bad -> OK
        if (param->state == Active) {
            on_ret(param);
        }

        id_not_awaited_now(id);

        break;              //state=Active

    default:
        return 1;
    }                       // states
    return 0;
}


// parameter is failed
void process_bad_value(Ant_mon_sensor * param)
{

    if((param->state == Fail_min) ||
       (param->state == Fail_max) ||
       (param->state == Fail_dep))
       return;

    // is param bad now?
    if (++(param->bad_count_cur) > param->bad_count) {

        if (param->value < param->min_value)
            param->state = Fail_min;
        else
            param->state = Fail_max;

        // reset counter
        param->bad_count_cur = 0;
    }
}

// parameter is returned to good
void process_good_value(Ant_mon_sensor * param)
{

    if (++(param->ret_count_cur) > param->ret_count) {

        param->state = Active;

        // reset counter
        param->ret_count_cur = 0;
    }
}

// parameter timed out
void process_timed_out(Ant_mon_sensor * param)
{

    if (param->state == Fail_tmout)
        return;

    param->state = Fail_tmout;

    param->ret_count_cur = 0;
    param->bad_count_cur = 0;
    param->next_check = global_time + param->tmout_interval;
    param->data_sent = 0;

    on_tmout(param);
}

// connection has failed
void process_conn_error(int i)
{
    int j, max;
    map < long int, void *>::iterator end, it;
    Ant_mon_sensor *p;

    // mark all its parameters failed...
    max = Sensors_by_ip[i]->size();
    end = params_by_id.end();
    for (j = 0; j < max; ++j) {
        p = (*(Sensors_by_ip[i]))[j];
        p->state = Fail_tmout;

        // mark it as not awaited
        for (it = params_by_id.begin(); it != end; ++it) {
            if ((Ant_mon_sensor *) (it->second) == p) {
                id_not_awaited_now(it->first);
                break;
            }
        }

        // make an action...
        on_dep((*(Sensors_by_ip[i]))[j]);
    }

    mark_dependent(Conn_by_ip[i], Fail_dep);

    on_conn_error(i);
}

// connection was established
void process_connected(int i)
{
    int j, max;
    map < long int, void *>::iterator end, it;
    Ant_mon_sensor *p;

    // mark all its parameters as good...
    max = Sensors_by_ip[i]->size();
    end = params_by_id.end();
    for (j = 0; j < max; ++j) {
        p = (*(Sensors_by_ip[i]))[j];
        p->state = Active;

        // mark it as unresolved
        p->id = -1;

        // make an action...
        //on_connect((*(Sensors_by_ip[i]))[j]);
    }

    was_change = 1;

    // was disconnected? (Not first connect)
    if (Conn_by_ip[i]->disconnected_time > 0)
        on_connect(i);

    Conn_by_ip[i]->disconnected_time = 0;
    Conn_by_ip[i]->error = 0;
}

// action: connection was failed (works NOT for sensor - for host only)
void on_conn_error(int i)
{
    Ant_action *act;
    const char *str;

    if (on_conn_err.size() < 1)
        return;

    // execute actions
    list < Ant_action * >::iterator end = on_conn_err.end();
    for (list < Ant_action * >::iterator it = on_conn_err.begin(); it != end;
         ++it) {
        act = *it;

        // substitute variables
        fill_vars_conn(&Variables, Conn_by_ip[i]);
        str = subst(act->argument.c_str(), Variables);

        act->sendMessage(str);
        act->flush();
    }
}

// action: connection was established (works NOT for sensor - for host only)
void on_connect(int i)
{
    Ant_action *act;
    const char *str;

    if (on_conn.size() < 1)
        return;

    // execute actions
    std::list < Ant_action * >::iterator end = on_conn.end();
    for (std::list < Ant_action * >::iterator it = on_conn.begin(); it != end;
         ++it) {
        act = *it;

        // substitute variables
        fill_vars_conn(&Variables, Conn_by_ip[i]);
        str = subst(act->argument.c_str(), Variables);

        act->sendMessage(str);
        act->flush();
    }
}

// action: parameter is good
void on_good(Ant_mon_sensor * param)
{
    Ant_action *act;
    const char *str;

    LOG(ALOG_SENS, ("!! OK %s on %s",
                    param->name.c_str(), param->conn->name.c_str()));

    if (param->on_good.size() < 1)
        return;

    // execute actions
    std::list < Ant_action * >::iterator end = param->on_good.end();
    for (std::list < Ant_action * >::iterator it = param->on_good.begin();
         it != end; ++it) {
        act = *it;

        // substitute variables
        fill_vars(&Variables, param);
        str = subst(act->argument.c_str(), Variables);

        act->sendMessage(str);
        act->flush();
    }
}

// action: parameter is bad
void on_bad(Ant_mon_sensor * param)
{
    Ant_action *act;
    const char *str;

    LOG(ALOG_SENS, ("!! NOT_OK %s on %s",
                    param->name.c_str(), param->conn->name.c_str()));

    if (param->on_bad.size() < 1)
        return;

    // execute actions
    std::list < Ant_action * >::iterator end = param->on_bad.end();
    for (std::list < Ant_action * >::iterator it = param->on_bad.begin();
         it != end; ++it) {
        act = *it;

        // substitute variables
        fill_vars(&Variables, param);
        str = subst(act->argument.c_str(), Variables);

        act->sendMessage(str);
        act->flush();
    }
}

// action: parameter was good and is bad now
void on_fail(Ant_mon_sensor * param)
{
    Ant_action *act;
    const char *str;

    LOG(ALOG_SENS, ("!! FAIL %s on %s",
                    param->name.c_str(), param->conn->name.c_str()));

    if (param->on_fail.size() < 1)
        return;

    // execute actions
    std::list < Ant_action * >::iterator end = param->on_fail.end();
    for (std::list < Ant_action * >::iterator it = param->on_fail.begin();
         it != end; ++it) {
        act = *it;

        // substitute variables
        fill_vars(&Variables, param);
        str = subst(act->argument.c_str(), Variables);

        act->sendMessage(str);
        act->flush();
    }
}

// action: parameter was bad is good now
void on_ret(Ant_mon_sensor * param)
{
    Ant_action *act;
    const char *str;

    LOG(ALOG_SENS, ("!! GOOD %s on %s",
                    param->name.c_str(), param->conn->name.c_str()));

    if (param->on_ret.size() < 1)
        return;

    // execute actions
    std::list < Ant_action * >::iterator end = param->on_ret.end();
    for (std::list < Ant_action * >::iterator it = param->on_ret.begin();
         it != end; ++it) {
        act = *it;

        // substitute variables
        fill_vars(&Variables, param);
        str = subst(act->argument.c_str(), Variables);

        act->sendMessage(str);
        act->flush();
    }
}

// action: parameter is bad, but first good value is received
void on_single_good(Ant_mon_sensor * param)
{
    Ant_action *act;
    const char *str;

    LOG(ALOG_SENS, ("!! good1 %s on %s",
                    param->name.c_str(), param->conn->name.c_str()));

    if (param->on_single_good.size() < 1)
        return;

    // execute actions
    std::list < Ant_action * >::iterator end = param->on_single_good.end();
    for (std::list < Ant_action * >::iterator it =
         param->on_single_good.begin(); it != end; ++it) {
        act = *it;

        // substitute variables
        fill_vars(&Variables, param);
        str = subst(act->argument.c_str(), Variables);

        act->sendMessage(str);
        act->flush();
    }
}

// action: parameter is good, but first bdd value is received
void on_single_bad(Ant_mon_sensor * param)
{
    Ant_action *act;
    const char *str;

    LOG(ALOG_SENS, ("!! bad1 %s on %s",
                    param->name.c_str(), param->conn->name.c_str()));

    if (param->on_single_bad.size() < 1)
        return;

    // execute actions
    std::list < Ant_action * >::iterator end = param->on_single_bad.end();
    for (std::list < Ant_action * >::iterator it =
         param->on_single_bad.begin(); it != end; ++it) {
        act = *it;

        // substitute variables
        fill_vars(&Variables, param);
        str = subst(act->argument.c_str(), Variables);

        act->sendMessage(str);
        act->flush();
    }
}

// action: parameter timed out
void on_tmout(Ant_mon_sensor * param)
{
    Ant_action *act;
    const char *str;

    LOG(ALOG_SENS, ("!! tmout %s on %s",
                    param->name.c_str(), param->conn->name.c_str()));

    if (param->on_tmout.size() < 1)
        return;

    if (!param->conn->connected) {
        LOG(ALOG_DEBUG, ("Skip. Not connected."));
        return;
    }

    // execute actions
    std::list < Ant_action * >::iterator end = param->on_tmout.end();
    for (std::list < Ant_action * >::iterator it = param->on_tmout.begin();
         it != end; ++it) {
        act = *it;

        // substitute variables
        fill_vars(&Variables, param);
        str = subst(act->argument.c_str(), Variables);

        act->sendMessage(str);
        act->flush();
    }
}

// action: parameter cannot be checked (connection timed out or dependence)
void on_dep(Ant_mon_sensor * param)
{
    Ant_action *act;
    const char *str;

    LOG(ALOG_SENS, ("!! dep %s on %s",
                    param->name.c_str(), param->conn->name.c_str()));

    if (param->on_dep.size() < 1)
        return;

    // execute actions
    std::list < Ant_action * >::iterator end = param->on_dep.end();
    for (std::list < Ant_action * >::iterator it = param->on_dep.begin();
         it != end; ++it) {
        act = *it;

        // substitute variables
        fill_vars(&Variables, param);
        str = subst(act->argument.c_str(), Variables);

        act->sendMessage(str);
        act->flush();
    }
}

// action: any head server failed
void on_any_head_death(int i)
{
    Ant_action *act;
    const char *str;

    LOG(ALOG_SENS, ("!! any_head_death %s", heads[i]->name.c_str()));

    if (head_death_any.size() < 1)
        return;

    // execute actions
    std::list < Ant_action * >::iterator end = head_death_any.end();
    for (std::list < Ant_action * >::iterator it = head_death_any.begin();
         it != end; ++it) {
        act = *it;

        // substitute variables
        fill_vars_head(&Variables, heads[i]);
        str = subst(act->argument.c_str(), Variables);

        act->sendMessage(str);
        act->flush();
    }
}

// action: head server failed
void on_head_death(int i)
{
    Ant_action *act;
    const char *str;

    LOG(ALOG_SENS, ("!! head_death %s", heads[i]->name.c_str()));

    if (head_death.size() < 1)
        return;

    // execute actions
    std::list < Ant_action * >::iterator end = head_death.end();
    for (std::list < Ant_action * >::iterator it = head_death.begin();
         it != end; ++it) {
        act = *it;

        // substitute variables
        fill_vars_head(&Variables, heads[i]);
        str = subst(act->argument.c_str(), Variables);

        act->sendMessage(str);
        act->flush();
    }
}

// mark all dependent parameters with given state
void mark_dependent(Ant_connection * conn, Ant_mon_sensor_state state)
{
    LOG(ALOG_WARN, ("MARK DEP not implemented"));

}

// generate new almost unique id
long int new_id()
{
    if (++_id > MAX_ID) {
        _id = 1;
        return 1;
    }
    return _id;
}


Ant_mon_sensor::Ant_mon_sensor()
{
    conn = NULL;
    name="";
    is_own = 1;
    controlled = 'n';
    type=Ant_Sensor_Data;
    mask=Ant_Min_Mask|Ant_Max_Mask;

    //name = out_name = last_name = NULL;
    passive = 0;
    id = -1;

    disconnected_time = global_time;

    value = 0;
    min_value = max_value = ret_min_value = ret_max_value = 0;
    min_rate = max_rate = ret_min_rate = ret_max_rate = 0;

    bad_count = bad_count_cur = ret_count = ret_count_cur = 0;
    bad_rcount = bad_rcount_cur = ret_rcount = ret_rcount_cur = 0;

    old_state = state = Active;
    heads[0] = -1;
}

Ant_mon_sensor::~Ant_mon_sensor()
{
    /*if (conn) {
        delete conn;
    }
    */
    /*
    if (name) {
        delete name;
    }

    if (out_name) {
        delete out_name;
    }

    if (last_name) {
        delete last_name;
    }
    */
}

//
//  Fill variables with parameter-dependent values
//
void fill_vars(map < string, string > *vars, Ant_mon_sensor * p)
{

    (*vars)["name"] = p->name;
    (*vars)["last_name"] = p->last_name;
    (*vars)["out_name"] = p->out_name;
    switch (p->state) {
    case Active:
        (*vars)["state"] = "OK";
        (*vars)["num_state"] = "0";
        break;
    case Fail_min:
        (*vars)["state"] = "Failed";
        (*vars)["num_state"] = "1";
        break;
    case Fail_max:
        (*vars)["state"] = "Failed";
        (*vars)["num_state"] = "2";
        break;
    case Fail_min_rate:
        (*vars)["state"] = "Failed";
        (*vars)["num_state"] = "11";
        break;
    case Fail_max_rate:
        (*vars)["state"] = "Failed";
        (*vars)["num_state"] = "12";
        break;
    case Fail_dep:
        (*vars)["state"] = "Failed by dependency";
        (*vars)["num_state"] = "3";
        break;
    case Fail_tmout:
        (*vars)["state"] = "Timed out";
        (*vars)["num_state"] = "4";
        break;
    case Fail_get:
        (*vars)["state"] = "Failed to get value";
        (*vars)["num_state"] = "5";
        break;
    case Disabled:
        (*vars)["state"] = "Disabled";
        (*vars)["num_state"] = "10";
        break;
    default:
        (*vars)["state"] = "Undefined";
        (*vars)["num_state"] = "100";
    }

    sprintf(char_buf, "%ld", p->id);
    (*vars)["id"] = char_buf;
    if(p->type==Ant_Sensor_String){
        (*vars)["value"] = p->strvalue;
    }
    else{
        sprintf(char_buf, "%f", p->value);
        (*vars)["value"] = char_buf;
    }
    sprintf(char_buf, "%f", p->min_value);
    (*vars)["min"] = char_buf;
    sprintf(char_buf, "%f", p->max_value);
    (*vars)["max"] = char_buf;
    sprintf(char_buf, "%f", p->ret_min_value);
    (*vars)["min_ret"] = char_buf;
    sprintf(char_buf, "%f", p->ret_max_value);
    (*vars)["max_ret"] = char_buf;
    sprintf(char_buf, "%d", p->bad_count);
    (*vars)["bad_count"] = char_buf;
    sprintf(char_buf, "%d", p->bad_count_cur);
    (*vars)["bad_count_cur"] = char_buf;
    sprintf(char_buf, "%d", p->ret_count);
    (*vars)["ret_count"] = char_buf;
    sprintf(char_buf, "%d", p->ret_count_cur);
    (*vars)["ret_count_cur"] = char_buf;
    sprintf(char_buf, "%ld", (long)p->interval);

    sprintf(char_buf, "%f", p->min_rate);
    (*vars)["min_rate"] = char_buf;
    sprintf(char_buf, "%f", p->max_rate);
    (*vars)["max_rate"] = char_buf;
    sprintf(char_buf, "%f", p->ret_min_rate);
    (*vars)["min_ret_rate"] = char_buf;
    sprintf(char_buf, "%f", p->ret_max_rate);
    (*vars)["max_ret_rate"] = char_buf;
    sprintf(char_buf, "%d", p->bad_rcount);
    (*vars)["bad_rcount"] = char_buf;
    sprintf(char_buf, "%d", p->bad_rcount_cur);
    (*vars)["bad_rcount_cur"] = char_buf;
    sprintf(char_buf, "%d", p->ret_rcount);
    (*vars)["ret_rcount"] = char_buf;
    sprintf(char_buf, "%d", p->ret_rcount_cur);
    (*vars)["ret_rcount_cur"] = char_buf;

    sprintf(char_buf, "%ld", (long)p->interval);
    (*vars)["interval"] = char_buf;
    sprintf(char_buf, "%ld", (long)p->timeout);
    (*vars)["timeout"] = char_buf;
    sprintf(char_buf, "%ld", (long)global_time);
    (*vars)["time"] = char_buf;
    (*vars)["date"] = timestamp();

    if (p->conn)
        (*vars)["addr"] = p->conn->name;
    else
        (*vars)["addr"] = string("*") + p->name + string("*");
}

//
//  Fill variables with head server dependent values
//
void fill_vars_head(map < string, string > *vars, Ant_head * p)
{

    (*vars)["name"] = p->name;
    (*vars)["out_name"] = p->name;
    (*vars)["addr"] = p->name;
    switch (p->state) {
    case Active:
        (*vars)["state"] = "OK";
        (*vars)["num_state"] = "0";
        break;
    case Fail_tmout:
        (*vars)["state"] = "Timed out";
        (*vars)["num_state"] = "1";
        break;
    case Disabled:
        (*vars)["state"] = "Disabled";
        (*vars)["num_state"] = "2";
        break;
    default:
        (*vars)["state"] = "Undefined";
        (*vars)["num_state"] = "100";
    }

    sprintf(char_buf, "%ld", (long)head_ping);
    (*vars)["interval"] = char_buf;
    sprintf(char_buf, "%ld", (long)head_tmout);
    (*vars)["timeout"] = char_buf;
    sprintf(char_buf, "%ld", (long)global_time);
    (*vars)["time"] = char_buf;
    (*vars)["date"] = timestamp();

/*  (*vars)["addr"]=p->out_conn->name; */
}

//
//  Fill variables with parameter-dependent values for Ant_connection object
//
void fill_vars_conn(map < string, string > *vars, Ant_connection * p)
{

    (*vars)["name"] = p->name;
    (*vars)["out_name"] = p->name;
    (*vars)["addr"] = p->name;
    if (p->connected) {
        (*vars)["state"] = "OK";
        (*vars)["num_state"] = "0";
    }
    else {
        (*vars)["state"] = "Disconnected";
        (*vars)["num_state"] = "1";
    }

    (*vars)["enabled"] = (p->enabled ? "1" : "0");
    sprintf(char_buf, "%ld", (long)global_time);
    (*vars)["time"] = char_buf;
    (*vars)["date"] = timestamp();
}

void daemonize()
{
    int pid;

    //if our parent isn't 'init' we do nothing
    if (getppid() != 1) {
        //claim signals ignoring
        signal(SIGTTOU, SIG_IGN);
        signal(SIGTTIN, SIG_IGN);
        signal(SIGTSTP, SIG_IGN);

        //create group and session
        if ((pid = fork())!=0) {
            //parent

            if (pid < 0) {
                LOG(ALOG_WARN, ("Cannot fork to daemonize"));
            }
            gentle_exit(0);
        }
        //child
        setsid();

    }
}

void process_mon_message(Ant_connection * conn)
{
    long int id;
    int i = 0;

    // read id
    id = 0;
    for (i = 0; conn->buffer[i] >= '0' && conn->buffer[i] <= '9'; ++i) {
        id = id * 10 + conn->buffer[i] - '0';
    }

    // check format
    if (conn->buffer[i] != ':' || conn->buffer[i + 2] != ':') {
        LOG(ALOG_WARN | ALOG_NET,
            ("Bad message from %s (%s)", conn->name.c_str(), conn->buffer));
        return;
    }
    ++i;

    // read action
    switch (conn->buffer[i]) {
        // heartbeat
    case 'h':
        heads[i]->last_ping = global_time;
        LOG(ALOG_DEBUG2,
            ("PING fron %s", conn->name.c_str()));
        break;

        // 'send me list of sensors, which i can support too'
    case 'l':
    default:
        LOG(ALOG_DEBUG,
            ("Not implemented intermonitor communication (%s)",
             conn->buffer));
        break;
    }

}


void config_reload(){
    long i, j, max, max2;
    long srch_i, srch_j, srch_max1, srch_max2;

    vector< vector<class Ant_mon_sensor * > *> saved_sensors;
    vector < class Ant_mon_sensor * > *v, *v2, *srch_v;
    Ant_mon_sensor *ptr, *sens;
    int found;

    /* close all agent connections */
    max=Conn_by_ip.size();

    for (i=0;i<max;++i){
        delete Conn_by_ip[i];
    }
    Conn_by_ip.clear();

    /* close all mons connections */
    max=heads.size();
    for (i=0;i<max;++i){
        Ant_head *hh=heads[i];
           for(j=0;j<hh->out_conn_list.size();++j){
               delete hh->out_conn_list[j];
           }

           for(j=0;j<hh->in_conn_list.size();++j){
               delete hh->in_conn_list[j];
           }
        delete hh;
    }
    heads.clear();

    /* actions cache save? */

    /* save sensor states */
    max=Sensors_by_ip.size();
    for (i=0;i<max;++i){
        v=new vector < class Ant_mon_sensor * >;
           saved_sensors.push_back(v);
        max2=Sensors_by_ip[i]->size();
        v2  =Sensors_by_ip[i];
        for(j=0;j<max2;++j){
            ptr=new Ant_mon_sensor;
            v->push_back(ptr);
            sens=(*v2)[j];
            ptr->bad_count_cur=sens->bad_count_cur;
            ptr->controlled=sens->controlled;
            ptr->id=sens->id;
            ptr->name=sens->name;
            if(sens->conn){
                ptr->last_name=sens->conn->name;
            }
            else{
                ptr->last_name="";
            }
            /* save connection name here */
            ptr->state=sens->state;
            ptr->old_state=sens->old_state;
            ptr->next_check=sens->next_check;
            ptr->ret_count_cur=sens->ret_count_cur;

            delete sens;
        }
        v2->clear();
    }

    /* clear all */
    Sensors_by_ip.clear();

    {
        vector < class Ant_connection * >::iterator  it0,e0;
        list < Ant_action * >::iterator              it1,e1;
        map < string, class Ant_action * >::iterator it2,e2;
        vector < class Ant_head * >::iterator        it4,e4;

        e0=Conn_by_ip.end();
        for(it0=Conn_by_ip.begin();it0!=e0;++it0){
            delete *it0;
        }
        Conn_by_ip.clear();

        on_conn.clear();

        on_conn_err.clear();

        e2=Actions.end();
        for(it2=Actions.begin();it2!=e2;++it2){
            delete (*it2).second;
        }
        Actions.clear();

        head_death.clear();

        head_death_any.clear();

        e4=heads.end();
        for(it4=heads.begin();it4!=e4;++it4){
            delete *it4;
        }
        heads.clear();

    }

    if (load_config("/etc/antmon.conf"))
        if (load_config("./antmon.conf")) {
            LOG(ALOG_WARN, ("Cannot open file antmon.conf"));
            gentle_exit(1);
        }
    /* restore sensors state */
    /* max,max2,v2,ptr - saved data */
    /* srch_*          - loaded sensors */
    max=saved_sensors.size();
    for (i=0;i<max;++i){
        max2=saved_sensors[i]->size();
        v2  =saved_sensors[i];
        for(j=0;j<max2;++j){
            ptr=(*v2)[j];

            /* try to search the same sensor */
            found=0;
            srch_max1=Sensors_by_ip.size();
            for(srch_i=0;srch_i<srch_max1;++srch_i){
                srch_v   =Sensors_by_ip[i];
                srch_max2=srch_v->size();
                for(srch_j=0;srch_j<srch_max2;++srch_j){
                    if((*srch_v)[srch_j]->name == ptr->name){
                        if((*srch_v)[srch_j]->conn!=NULL){
                            if((*srch_v)[srch_j]->conn->name == ptr->last_name){
                                /* YES! It maches! */
                                found=1;
                                break;
                            }
                        }
                        else{
                            found=1;
                            break;
                        }
                    }
                }
                if(found)
                    break;
            }
            if(found){
                /* restore saved data! */
                /* v[srch_j] = current sensor, ptr = saved data */
                (*srch_v)[srch_j]->bad_count_cur=ptr->bad_count_cur;
                (*srch_v)[srch_j]->controlled=ptr->controlled;
                (*srch_v)[srch_j]->id=ptr->id;
                (*srch_v)[srch_j]->state=ptr->state;
                (*srch_v)[srch_j]->old_state=ptr->old_state;
                (*srch_v)[srch_j]->next_check=ptr->next_check;
                (*srch_v)[srch_j]->ret_count_cur=ptr->ret_count_cur;
            } /* found */

            delete ptr;
        } /* for all sensors */
        delete v2;
    } /* for all saved data */
}

void create_pid(void){
    FILE *f;
    long int pid;

    f=fopen("/var/run/ant_mon.pid","r");
    if(f!=NULL){
        if(fscanf(f,"%ld",&pid)>0){
            if(kill(pid,0)==0){
                LOG(ALOG_ALL, ("AntMon server is already running with pid %ld. Exit",pid));
                gentle_exit(1);
            }
        }
        fclose(f);
    }

    f=fopen("/var/run/ant_mon.pid","w");
    if(f!=NULL){
        fprintf(f,"%ld",(long)getpid());
        //printf("My PID is %d\n",getpid());
        fclose(f);
    }
    else{
        LOG(ALOG_WARN, ("Cannot create /var/run/ant_mon.pid"));
    }
}

void gentle_exit(int code){
    kill(0, SIGQUIT);
    unlink("/var/run/ant_mon.pid");
    exit(code);
}
