#ifndef ANT_MON_HPP
#define ANT_MON_HPP

#include <sys/types.h>
#include <list>

#include "ant_conn.hpp"
#include "ant_act.hpp"

#define ANT_MAX_HEADS 16
#define ANT_HEAD_PORT 9989
#define ANT_PORT 8898


enum Ant_mon_sensor_state {Active=0,
                           Fail_min,
                           Fail_max,
                           Fail_min_rate,
                           Fail_max_rate,
                           Fail_dep,
                           Fail_tmout,
                           Fail_get,
                           Disabled};

enum Ant_sensor_type {Ant_Sensor_Data,
                      Ant_Sensor_String,
                      Ant_Sensor_Counter,
                     };

typedef unsigned int Ant_sensor_mask;

#define Ant_Min_Mask 1
#define Ant_Max_Mask 2
#define Ant_Min_Rate_Mask 4
#define Ant_Max_Rate_Mask 8

class Ant_mon_sensor{
  //  Ant_mon_sensor(Ant_mon_sensor&){}

public:
  Ant_connection *conn;

  //  Ant_connection *mon; //if extern monitor

  string        name;
  string        last_name;
  string        out_name;
  char          passive;
  char          is_own;     // controlled by me
  char          controlled; // we are backup monitor now
  long int      id;

  Ant_sensor_mask mask;
  Ant_sensor_type type;

  float         value;
  float         min_value;
  float         max_value;
  float         ret_min_value;
  float         ret_max_value;

  float         rate;
  float         min_rate;
  float         max_rate;
  float         ret_min_rate;
  float         ret_max_rate;

  short int     heads[ANT_MAX_HEADS];

  // how much times value may be bad legally
  int           bad_count;
  // how much times value is bad already
  int           bad_count_cur;

  // how much times value must be good to be marked good
  int           ret_count;
  // how much times value is good really
  int           ret_count_cur;

  // as above for rate
  int           bad_rcount;
  int           bad_rcount_cur;
  int           ret_rcount;
  int           ret_rcount_cur;

  string strvalue;
  string stretalon;

  time_t        interval;          // data refresh interval
  time_t        tmout_interval;    // --"-----"--- when timed out
  time_t        disconnected_time; // when was disconnection
  time_t        timeout;           // data refresh timeout (MUST be < interval)
  time_t        data_sent;         // when data request was sent
  time_t        next_check;        // next check time
  time_t        last_time;         // previous data retrival time

  std::list<Ant_mon_sensor *> depends;
  std::list<Ant_action *> on_ret;         // returned to good
  std::list<Ant_action *> on_good;        // is good
  std::list<Ant_action *> on_single_good; // is bad, but single good received
  std::list<Ant_action *> on_fail;        // failed
  std::list<Ant_action *> on_bad;         // is bad
  std::list<Ant_action *> on_single_bad;  // is good, but single bad received
  std::list<Ant_action *> on_tmout;       // timed out
  std::list<Ant_action *> on_dep;         // failed by dependence

  Ant_mon_sensor_state state;
  Ant_mon_sensor_state old_state;

  Ant_mon_sensor();
  ~Ant_mon_sensor();
};

enum Ant_head_state {Head_Active=0,Head_Fail,Head_Disabled,Head_No_Init};

class Ant_head{
public:
  std::string    name;
  std::list<std::string *> ips;

  Ant_connection *out_conn;
  Ant_connection *in_conn;

  vector<Ant_connection *> out_conn_list;
  vector<Ant_connection *> in_conn_list;

  Ant_head_state state;
  int            timed_out; //flag = action was made
  time_t         last_ping;
  time_t         next_ping;
};

#endif
